//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#ifndef FACTOR_IMU2D_THETA_H_
#define FACTOR_IMU2D_THETA_H_

//Wolf includes
#include "imu/feature/feature_imu2d.h"
#include "imu/processor/processor_imu2d.h"
#include "imu/sensor/sensor_imu2d.h"
#include "core/factor/factor_autodiff.h"
#include "core/math/rotations.h"
#include "imu/math/imu2d_tools.h"

//Eigen include

namespace wolf {
    
WOLF_PTR_TYPEDEFS(FactorImu2d);

//class
class FactorImu2d : public FactorAutodiff<FactorImu2d, 8, 2, 1, 2, 3, 2, 1, 2, 3>
{
    public:
        FactorImu2d(const FeatureImu2dPtr& _ftr_ptr,
                    const CaptureImuPtr& _capture_origin_ptr,
                    const ProcessorBasePtr& _processor_ptr,
                    bool _apply_loss_function,
                    FactorStatus _status = FAC_ACTIVE);

        ~FactorImu2d() override = default;

        /** \brief : compute the residual from the state blocks being iterated by the solver.
            -> computes the expected measurement
            -> corrects actual measurement with new bias
            -> compares the corrected measurement with the expected one
            -> weights the result with the covariance of the noise (residual = sqrt_info_matrix * err;)
        */
        template<typename T>
        bool operator ()(const T* const _p1,
                         const T* const _o1,
                         const T* const _v1,
                         const T* const _b1,
                         const T* const _p2,
                         const T* const _o2,
                         const T* const _v2,
                         const T* const _b2,
                         T* _res) const;
        Eigen::Matrix3d getBiasDriftSquareRootInformationUpper(const FeatureImu2dPtr& _ftr_ptr)
        {
            Eigen::Matrix3d res = Eigen::Matrix3d::Identity();
            if(_ftr_ptr->getCapture()->getSensor()){
              res *= 1./(std::static_pointer_cast<SensorImu2d>(_ftr_ptr->getCapture()->getSensor())->getAbRateStdev() * dt_);
              res(2,2) = 1./(std::static_pointer_cast<SensorImu2d>(_ftr_ptr->getCapture()->getSensor())->getWbRateStdev() * dt_);
            }
            WOLF_WARN_COND(!_ftr_ptr->getCapture()->getSensor(), "Null Sensor Pointer");
            return res;
        }

    private:
        // Imu processor
        ProcessorImu2dPtr processor_imu2d_;

        /// Preintegrated delta
        Eigen::Vector5d delta_preint_;

        // Biases used during preintegration
        Eigen::Vector3d bias_preint_;

        // Jacobians of preintegrated deltas wrt biases
        Eigen::Matrix<double, 5, 3> jacobian_bias_;

        /// Metrics
        const double dt_; ///< delta-time and delta-time-squared between keyframes
        
        /** bias covariance and bias residuals
         *
         * continuous bias covariance matrix for accelerometer would then be A_r = diag(ab_stdev_^2, ab_stdev_^2, ab_stdev_^2)
         * To compute bias residuals, we will need to do (sqrt(A_r)).inverse() * ab_error
         *
         * In our case, we introduce time 'distance' in the computation of this residual (SEE FORSTER17), thus we have to use the discret covariance matrix
         * discrete bias covariance matrix for accelerometer : A_r_dt = dt_ * A_r
         * taking the square root for bias residuals : sqrt_A_r_dt = sqrt(dt_ * A_r) = sqrt(dt_) * sqrt(A_r)
         * then with the inverse : sqrt_A_r_dt_inv = (sqrt(dt_ * A_r)).inverse() = (1/sqrt(dt_)) * sqrt(A_r).inverse()
         *
         * same logic for gyroscope bias
         */ 
        const Eigen::Matrix5d sqrt_delta_preint_inv_;
        const Eigen::Matrix3d sqrt_bias_drift_dt_inv_;

    public:
      EIGEN_MAKE_ALIGNED_OPERATOR_NEW;
};

///////////////////// IMPLEMENTAITON ////////////////////////////

inline FactorImu2d::FactorImu2d(const FeatureImu2dPtr&    _ftr_ptr,
                                const CaptureImuPtr&    _cap_origin_ptr,
                                const ProcessorBasePtr& _processor_ptr,
                                bool                    _apply_loss_function,
                                FactorStatus        _status) :
                FactorAutodiff<FactorImu2d, 8, 2, 1, 2, 3, 2, 1, 2, 3>( // ...
                        "FactorImu2d",
                        TOP_MOTION,
                        _ftr_ptr,
                        _cap_origin_ptr->getFrame(),
                        _cap_origin_ptr,
                        nullptr,
                        nullptr,
                        _processor_ptr,
                        _apply_loss_function,
                        _status,
                        _cap_origin_ptr->getFrame()->getP(),
                        _cap_origin_ptr->getFrame()->getO(),
                        _cap_origin_ptr->getFrame()->getV(),
                        _cap_origin_ptr->getSensorIntrinsic(),
                        _ftr_ptr->getFrame()->getP(),
                        _ftr_ptr->getFrame()->getO(),
                        _ftr_ptr->getFrame()->getV(),
                        _ftr_ptr->getCapture()->getSensorIntrinsic()),
        processor_imu2d_(std::static_pointer_cast<ProcessorImu2d>(_processor_ptr)),
        delta_preint_(_ftr_ptr->getMeasurement()), // dp, dth, dv at preintegration time
        bias_preint_(_ftr_ptr->getBiasPreint()), // state biases at preintegration time
        jacobian_bias_(_ftr_ptr->getJacobianBias()), // Jacs of dp dv dq wrt biases
        dt_(_ftr_ptr->getFrame()->getTimeStamp() - _cap_origin_ptr->getTimeStamp()),
        sqrt_delta_preint_inv_(_ftr_ptr->getMeasurementSquareRootInformationUpper()),
        sqrt_bias_drift_dt_inv_(getBiasDriftSquareRootInformationUpper(_ftr_ptr))
{
    //
}

template<typename T>
inline bool FactorImu2d::operator ()(const T* const _p1,
                                   const T* const _th1,
                                   const T* const _v1,
                                   const T* const _b1,
                                   const T* const _p2,
                                   const T* const _th2,
                                   const T* const _v2,
                                   const T* const _b2,
                                   T* _res) const
{
    using namespace Eigen;

    // MAPS
    Map<const Matrix<T,2,1> > p1(_p1);
    const T& th1              = *_th1;
    Map<const Matrix<T,2,1> > v1(_v1);
    Map<const Matrix<T,3,1> > b1(_b1);

    Map<const Matrix<T,2,1> > p2(_p2);
    const T& th2              = *_th2;
    Map<const Matrix<T,2,1> > v2(_v2);
    Map<const Matrix<T,3,1> > b2(_b2);

    Map<Matrix<T,8,1> > res(_res);

    //residual
    /*
     * MATH:
     * res_delta = (Covariancia delta)^(T/2) * ((delta_preint (+) Jacob^delta_bias*(b1-b1_preint))- (x2 (-) x1)) 
     */
    Matrix<T, 5, 1> delta_step = jacobian_bias_*(b1 - bias_preint_);
    Matrix<T, 5, 1> delta_preint = delta_preint_.cast<T>();
    Matrix<T, 5, 1> delta_corr = imu2d::plus(delta_preint, delta_step);

    Matrix<T, 5, 1> delta_predicted;
    Map<Matrix<T,2,1> > dp_predicted( &delta_predicted(0) + 0);
    T& dth_predicted = delta_predicted(2);
    Map<Matrix<T,2,1> > dv_predicted(&delta_predicted(0) + 3);
    imu2d::between(p1, th1, v1, p2, th2, v2, dt_, dp_predicted, dth_predicted, dv_predicted); 

    Matrix<T, 5, 1> delta_error = delta_corr - delta_predicted;
    delta_error(2) = pi2pi(delta_error(2));

    res.template head<5>() = sqrt_delta_preint_inv_*delta_error;
    

    //bias drift
    res.template tail<3>() = sqrt_bias_drift_dt_inv_*(b2 -b1);

    return true;
}

} // namespace wolf

#endif
