//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#ifndef SENSOR_IMU2D_H
#define SENSOR_IMU2D_H

//wolf includes
#include "core/sensor/sensor_base.h"
#include "core/utils/params_server.h"
#include <iostream>

namespace wolf {

WOLF_STRUCT_PTR_TYPEDEFS(ParamsSensorImu2d);


struct ParamsSensorImu2d : public ParamsSensorBase
{
    //noise std dev
    double w_noise = 0.001; //standard deviation of Gyroscope noise (same for all the axis) in rad/sec/ sqrt(s)
    double a_noise = 0.04; //standard deviation of Acceleration noise (same for all the axis) in m/s2/sqrt(s)

    //Initial biases std dev
    double ab_initial_stdev = 0.01; //accelerometer micro_g/sec
    double wb_initial_stdev = 0.01; //gyroscope rad/sec

    // bias rate of change std dev
    double ab_rate_stdev = 0.00001;
    double wb_rate_stdev = 0.00001;

    ~ParamsSensorImu2d() override = default;
    ParamsSensorImu2d()
    {
        //DEFINED FOR COMPATIBILITY PURPOSES. TO BE REMOVED IN THE FUTURE.
    }
    ParamsSensorImu2d(std::string _unique_name, const ParamsServer& _server):
        ParamsSensorBase(_unique_name, _server)
    {
        w_noise             = _server.getParam<double>(prefix + _unique_name + "/w_noise");
        a_noise             = _server.getParam<double>(prefix + _unique_name + "/a_noise");
        ab_initial_stdev    = _server.getParam<double>(prefix + _unique_name + "/ab_initial_stdev");
        wb_initial_stdev    = _server.getParam<double>(prefix + _unique_name + "/wb_initial_stdev");
        ab_rate_stdev       = _server.getParam<double>(prefix + _unique_name + "/ab_rate_stdev");
        wb_rate_stdev       = _server.getParam<double>(prefix + _unique_name + "/wb_rate_stdev");
    }
    std::string print() const override
    {
        return ParamsSensorBase::print()                           + "\n"
            + "w_noise: "           + std::to_string(w_noise)           + "\n"
            + "a_noise: "           + std::to_string(a_noise)           + "\n"
            + "ab_initial_stdev: "  + std::to_string(ab_initial_stdev)  + "\n"
            + "wb_initial_stdev: "  + std::to_string(wb_initial_stdev)  + "\n"
            + "ab_rate_stdev: "     + std::to_string(ab_rate_stdev)     + "\n"
            + "wb_rate_stdev: "     + std::to_string(wb_rate_stdev)     + "\n";
    }
};

WOLF_PTR_TYPEDEFS(SensorImu2d);

class SensorImu2d : public SensorBase
{

    protected:
        double a_noise; //Power Spectral Density (same for all the axis) in micro_g/ sqrt(Hz)
        double w_noise; //Rate Noise Spectral Density (same for all the axis) in deg/sec/ sqrt(Hz)

        //This is a trial to factor how much can the bias change in 1 sec at most
        double ab_initial_stdev; //accelerometer m/sec^s
        double wb_initial_stdev; //gyroscope rad/sec
        double ab_rate_stdev;    //accelerometer m/sec^2 / sqrt(sec)
        double wb_rate_stdev;    //gyroscope rad/sec / sqrt(sec)

    public:

        SensorImu2d(const Eigen::VectorXd& _extrinsics, const ParamsSensorImu2d& _params);
        SensorImu2d(const Eigen::VectorXd& _extrinsics, ParamsSensorImu2dPtr _params);
        WOLF_SENSOR_CREATE(SensorImu2d, ParamsSensorImu2d, 3);

        double getGyroNoise() const;
        double getAccelNoise() const;
        double getWbInitialStdev() const;
        double getAbInitialStdev() const;
        double getWbRateStdev() const;
        double getAbRateStdev() const;

        ~SensorImu2d() override;

};

inline double SensorImu2d::getGyroNoise() const
{
    return w_noise;
}

inline double SensorImu2d::getAccelNoise() const
{
    return a_noise;
}

inline double SensorImu2d::getWbInitialStdev() const
{
    return wb_initial_stdev;
}

inline double SensorImu2d::getAbInitialStdev() const
{
    return ab_initial_stdev;
}

inline double SensorImu2d::getWbRateStdev() const
{
    return wb_rate_stdev;
}

inline double SensorImu2d::getAbRateStdev() const
{
    return ab_rate_stdev;
}

} // namespace wolf

#endif // SENSOR_Imu2D_H
