//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include "imu/processor/processor_compass.h"
#include "imu/capture/capture_compass.h"
#include "imu/factor/factor_compass_3d.h"

namespace wolf {

ProcessorCompass::ProcessorCompass(ParamsProcessorCompassPtr _params) :
        ProcessorBase("ProcessorCompass", 3, _params)
{
    //
}

ProcessorCompass::~ProcessorCompass()
{
    //
}

void ProcessorCompass::processCapture(CaptureBasePtr _capture)
{
    // Search for any stored frame within time tolerance of capture
    auto keyframe = buffer_frame_.select(_capture->getTimeStamp(), params_->time_tolerance);
    if (keyframe)
    {
        processMatch(keyframe, _capture);

        // remove the frame and older frames
        buffer_frame_.removeUpTo(keyframe->getTimeStamp());
    }
    // Otherwise: store capture
    // Note that more than one processor can be emplacing frames, so an older frame can arrive later than this one.
    else
        buffer_capture_.emplace(_capture->getTimeStamp(), _capture);
}

void ProcessorCompass::processKeyFrame(FrameBasePtr _frame)
{
    // Search for any stored capture within time tolerance of frame
    auto capture = buffer_capture_.select(_frame->getTimeStamp(), getTimeTolerance());
    if (capture)
    {
        processMatch(_frame, capture);

        // remove (very) old captures
        buffer_capture_.removeUpTo(_frame->getTimeStamp() - 10);
    }
    // Otherwise: If frame is more recent than any capture in buffer -> store frame to be processed later in processCapture
    else if (buffer_capture_.selectLastAfter(_frame->getTimeStamp(), getTimeTolerance()) == nullptr)
    {
        // store frame
        buffer_frame_.emplace(_frame->getTimeStamp(), _frame);
    }
    // Otherwise: There are captures more recent than the frame but none that match with it -> discard frame
}

void ProcessorCompass::processMatch(FrameBasePtr _frame, CaptureBasePtr _capture)
{
    auto cap_compass = std::dynamic_pointer_cast<CaptureCompass>(_capture);
    assert(cap_compass && "ProcessorCompass::processMatch capture not of type CaptureCompass");

    // link capture to the frame
    _capture->link(_frame);

    // emplace feature
    auto feat = FeatureBase::emplace<FeatureBase>(cap_compass,
                                                  "FeatureCompass",
                                                  cap_compass->getData(),
                                                  cap_compass->getDataCovariance());

    // emplace factor
    FactorBase::emplace<FactorCompass3d>(feat,
                                         feat,
                                         shared_from_this(),
                                         params_->apply_loss_function);
}

} /* namespace wolf */

// Register in the FactorySensor
#include "core/processor/factory_processor.h"

namespace wolf
{
WOLF_REGISTER_PROCESSOR(ProcessorCompass)
WOLF_REGISTER_PROCESSOR_AUTO(ProcessorCompass)
}

