//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
// imu
#include "imu/processor/processor_imu.h"
#include "imu/factor/factor_imu.h"
#include "imu/factor/factor_imu_bias_static.h"
#include "imu/math/imu_tools.h"

// wolf
#include <core/state_block/state_composite.h>

namespace wolf {

ProcessorImu::ProcessorImu(ParamsProcessorImuPtr _params_motion_imu) :
        ProcessorMotion("ProcessorImu", "POV", 3, 10, 10, 9, 6, 6, _params_motion_imu),
        params_motion_Imu_(std::make_shared<ParamsProcessorImu>(*_params_motion_imu))
{
    //
}

ProcessorImu::~ProcessorImu()
{
    //
}
void ProcessorImu::preProcess()
{
    auto cap_ptr = std::dynamic_pointer_cast<CaptureImu>(incoming_ptr_);
    assert(cap_ptr != nullptr && ("Capture type mismatch. Processor " + getName() + " can only process captures of type CaptureImu").c_str());
}
bool ProcessorImu::voteForKeyFrame() const
{
    // time span
    if (getBuffer().back().ts_ - getBuffer().front().ts_ > params_motion_Imu_->max_time_span)
    {
        WOLF_DEBUG( "PM: vote: time span" );
        return true;
    }
    // buffer length
    if (getBuffer().size() > params_motion_Imu_->max_buff_length)
    {
        WOLF_DEBUG( "PM: vote: buffer length" );
        return true;
    }
    // angle turned
    double angle = 2.0 * asin(delta_integrated_.segment(3,3).norm());
    if (angle > params_motion_Imu_->angle_turned)
    {
        WOLF_DEBUG( "PM: vote: angle turned" );
        return true;
    }
    //WOLF_DEBUG( "PM: do not vote" );
    return false;
}


CaptureMotionPtr ProcessorImu::emplaceCapture(const FrameBasePtr& _frame_own,
                                              const SensorBasePtr& _sensor,
                                              const TimeStamp& _ts,
                                              const VectorXd& _data,
                                              const MatrixXd& _data_cov,
                                              const VectorXd& _calib,
                                              const VectorXd& _calib_preint,
                                              const CaptureBasePtr& _capture_origin)
{
    auto cap_motion = std::static_pointer_cast<CaptureMotion>(CaptureBase::emplace<CaptureImu>(_frame_own,
                                                                                               _ts,
                                                                                               _sensor,
                                                                                               _data,
                                                                                               _data_cov,
                                                                                               _capture_origin));
    if (getSensor()->isStateBlockDynamic('I'))
        setCalibration(cap_motion, _calib);
    cap_motion->setCalibrationPreint(_calib_preint);

    return cap_motion;
}

FeatureBasePtr ProcessorImu::emplaceFeature(CaptureMotionPtr _capture_motion)
{
    auto feature = FeatureBase::emplace<FeatureImu>(_capture_motion,
                                                    _capture_motion->getBuffer().back().delta_integr_,
                                                    _capture_motion->getBuffer().back().delta_integr_cov_ + unmeasured_perturbation_cov_,
                                                    _capture_motion->getCalibrationPreint(),
                                                    _capture_motion->getBuffer().back().jacobian_calib_);
    return feature;
}

VectorXd ProcessorImu::getCalibration (const CaptureBasePtr _capture) const
{
    if (_capture)
        return _capture->getStateBlock('I')->getState();
    else
        return getSensor()->getStateBlockDynamic('I')->getState();
}

void ProcessorImu::setCalibration (const CaptureBasePtr _capture, const VectorXd& _calibration)
{
    _capture->getSensorIntrinsic()->setState(_calibration);
}

FactorBasePtr ProcessorImu::emplaceFactor(FeatureBasePtr _feature_motion, CaptureBasePtr _capture_origin)
{
    assert(_capture_origin != nullptr && _capture_origin->getType() == "CaptureImu" && "ProcessorImu::emplaceFactor: capture is null or not of type CaptureImu");
    assert(_feature_motion != nullptr && _feature_motion->getType() == "FeatureImu" && "ProcessorImu::emplaceFactor: feature is null or not of type FeatureImu");
    assert(_capture_origin->getFrame() && "ProcessorImu::emplaceFactor: capture not linked to a frame");
    assert(_feature_motion->getCapture() and _feature_motion->getCapture()->getFrame() && "ProcessorImu::emplaceFactor: feature not linked to a capture or a frame");

    CaptureImuPtr cap_imu = std::static_pointer_cast<CaptureImu>(_capture_origin);
    FeatureImuPtr ftr_imu = std::static_pointer_cast<FeatureImu>(_feature_motion);

    FactorBasePtr fac_imu;

    if (getSensor()->isStateBlockDynamic('I'))
        fac_imu = FactorBase::emplace<FactorImu>(_feature_motion,
                                                 ftr_imu,
                                                 cap_imu,
                                                 shared_from_this(),
                                                 params_->apply_loss_function);
    else
        fac_imu = FactorBase::emplace<FactorImuBiasStatic>(_feature_motion,
                                                           ftr_imu,
                                                           cap_imu,
                                                           shared_from_this(),
                                                           params_->apply_loss_function);

    return fac_imu;
}

void ProcessorImu::computeCurrentDelta(const Eigen::VectorXd& _data,
                                       const Eigen::MatrixXd& _data_cov,
                                       const Eigen::VectorXd& _calib,
                                       const double _dt,
                                       Eigen::VectorXd& _delta,
                                       Eigen::MatrixXd& _delta_cov,
                                       Eigen::MatrixXd& _jac_delta_calib) const
{
    assert(_data.size() == data_size_ && "Wrong data size!");

    using namespace Eigen;
    Matrix<double, 9, 6> jac_delta_data;

    /*
     * We have the following computing pipeline:
     *     Input: data, calib, dt
     *     Output: delta, delta_cov, jac_calib
     *
     * We do:
     *     body         = data - calib
     *     delta        = body2delta(body, dt) --> jac_body
     *     jac_data     = jac_body
     *     jac_calib    = - jac_body
     *     delta_cov  <-- propagate data_cov using jac_data
     *
     * where body2delta(.) produces a delta from body=(a,w) as follows:
     *     dp = 1/2 * a * dt^2
     *     dq = exp(w * dt)
     *     dv = a * dt
     */

    // create delta
    imu::body2delta(_data - _calib, _dt, _delta, jac_delta_data); // Jacobians tested in imu_tools

    // compute delta_cov
    _delta_cov = jac_delta_data * _data_cov * jac_delta_data.transpose();

    // compute jacobian_calib
    _jac_delta_calib = - jac_delta_data;

}

void ProcessorImu::deltaPlusDelta(const Eigen::VectorXd& _delta_preint,
                                  const Eigen::VectorXd& _delta,
                                  const double _dt,
                                  Eigen::VectorXd& _delta_preint_plus_delta) const
{
    /* MATHS according to Sola-16
     * Dp' = Dp + Dv*dt + 1/2*Dq*(a-a_b)*dt^2    = Dp + Dv*dt + Dq*dp   if  dp = 1/2*(a-a_b)*dt^2
     * Dv' = Dv + Dq*(a-a_b)*dt                  = Dv + Dq*dv           if  dv = (a-a_b)*dt
     * Dq' = Dq * exp((w-w_b)*dt)                = Dq * dq              if  dq = exp((w-w_b)*dt)
     *
     * where (dp, dq, dv) need to be computed in data2delta(), and Dq*dx =is_equivalent_to= Dq*dx*Dq'.
     */
    _delta_preint_plus_delta = imu::compose(_delta_preint, _delta, _dt);
}

void ProcessorImu::statePlusDelta(const VectorComposite& _x,
                                  const Eigen::VectorXd& _delta,
                                  const double _dt,
                                  VectorComposite& _x_plus_delta) const
{
    assert(_delta.size() == 10 && "Wrong _delta vector size");
    assert(_dt >= 0 && "Time interval _dt is negative!");

    const auto& delta = VectorComposite(_delta, "POV", {3,4,3});

    /* MATH according to Sola-16
     *
     * p' = p + v*dt + 0.5*g*dt^2 + q*dp
     * q' =                         q*dq
     * v' = v + g*dt +              q*dv
     */
    _x_plus_delta = imu::composeOverState(_x, delta, _dt);
}

void ProcessorImu::deltaPlusDelta(const Eigen::VectorXd& _delta_preint,
                                  const Eigen::VectorXd& _delta,
                                  const double _dt,
                                  Eigen::VectorXd& _delta_preint_plus_delta,
                                  Eigen::MatrixXd& _jacobian_delta_preint,
                                  Eigen::MatrixXd& _jacobian_delta) const
{
    /*
     * Expression of the delta integration step, D' = D (+) d:
     *
     *     Dp' = Dp + Dv*dt + Dq*dp
     *     Dv' = Dv + Dq*dv
     *     Dq' = Dq * dq
     *
     * Jacobians for covariance propagation.
     *
     * a. With respect to Delta, gives _jacobian_delta_preint = D'_D as:
     *
     *   D'_D = [ I    -DR*skew(dp)   I*dt
     *            0     dR.tr          0
     *            0    -DR*skew(dv)    I  ] // See Sola-16
     *
     * b. With respect to delta, gives _jacobian_delta = D'_d as:
     *
     *   D'_d = [ DR   0    0
     *            0    I    0
     *            0    0    DR ] // See Sola-16
     *
     * Note: covariance propagation, i.e.,  P+ = D_D * P * D_D' + D_d * M * D_d', is done in ProcessorMotion.
     */
    imu::compose(_delta_preint, _delta, _dt, _delta_preint_plus_delta, _jacobian_delta_preint, _jacobian_delta); // jacobians tested in imu_tools
}

Eigen::VectorXd ProcessorImu::correctDelta (const Eigen::VectorXd& delta_preint,
                                            const Eigen::VectorXd& delta_step) const
{
    return imu::plus(delta_preint, delta_step);
}



} // namespace wolf

// Register in the FactorySensor
#include "core/processor/factory_processor.h"

namespace wolf
{
WOLF_REGISTER_PROCESSOR(ProcessorImu)
WOLF_REGISTER_PROCESSOR_AUTO(ProcessorImu)
}
