//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
// imu
#include "imu/processor/processor_imu2d.h"
#include "imu/factor/factor_imu2d.h"
#include "imu/math/imu2d_tools.h"

// wolf
#include <core/state_block/state_composite.h>

namespace wolf {

  ProcessorImu2d::ProcessorImu2d(ParamsProcessorImu2dPtr _params_motion_imu) :
    ProcessorMotion("ProcessorImu2d", "POV", 2, 5, 5, 5, 6, 3, _params_motion_imu),
    params_motion_Imu_(std::make_shared<ParamsProcessorImu2d>(*_params_motion_imu))
  {
      //
  }

  ProcessorImu2d::~ProcessorImu2d()
  {
    //
  }

  void ProcessorImu2d::preProcess()
  {
    auto cap_ptr = std::dynamic_pointer_cast<CaptureImu>(incoming_ptr_);
    assert(cap_ptr != nullptr && ("Capture type mismatch. Processor " + getName() + " can only process captures of type CaptureImu").c_str());
  }

  bool ProcessorImu2d::voteForKeyFrame() const
  {
    // time span
    if (getBuffer().back().ts_ - getBuffer().front().ts_ > params_motion_Imu_->max_time_span)
    {
      WOLF_DEBUG( "PM: vote: time span" );
      return true;
    }
    // buffer length
    if (getBuffer().size() > params_motion_Imu_->max_buff_length)
    {
      WOLF_DEBUG( "PM: vote: buffer length" );
      return true;
    }
    // angle turned
    double angle = std::abs(delta_integrated_(2));
    if (angle > params_motion_Imu_->angle_turned)
    {
      WOLF_DEBUG( "PM: vote: angle turned" );
      return true;
    }
    //WOLF_DEBUG( "PM: do not vote" );
    return false;
  }


  CaptureMotionPtr ProcessorImu2d::emplaceCapture(const FrameBasePtr& _frame_own,
                                                  const SensorBasePtr& _sensor,
                                                  const TimeStamp& _ts,
                                                  const VectorXd& _data,
                                                  const MatrixXd& _data_cov,
                                                  const VectorXd& _calib,
                                                  const VectorXd& _calib_preint,
                                                  const CaptureBasePtr& _capture_origin)
  {
    auto cap_motion = std::static_pointer_cast<CaptureMotion>(CaptureBase::emplace<CaptureImu>(_frame_own,
                                                                                              _ts,
                                                                                              _sensor,
                                                                                              _data,
                                                                                              _data_cov,
                                                                                              _capture_origin));
    setCalibration(cap_motion, _calib);
    cap_motion->setCalibrationPreint(_calib_preint);

    return cap_motion;
  }

  FeatureBasePtr ProcessorImu2d::emplaceFeature(CaptureMotionPtr _capture_motion)
  {
    auto feature = FeatureBase::emplace<FeatureImu2d>(_capture_motion,
                                                      _capture_motion->getBuffer().back().delta_integr_,
                                                      _capture_motion->getBuffer().back().delta_integr_cov_ + unmeasured_perturbation_cov_,
                                                      _capture_motion->getCalibrationPreint(),
                                                      _capture_motion->getBuffer().back().jacobian_calib_);
    return feature;
  }

  VectorXd ProcessorImu2d::getCalibration (const CaptureBasePtr _capture) const
  {
      if (_capture)
          return _capture->getStateBlock('I')->getState();
      else
          return getSensor()->getStateBlockDynamic('I')->getState();
  }

  void ProcessorImu2d::setCalibration (const CaptureBasePtr _capture, const VectorXd& _calibration)
  {
    _capture->getSensorIntrinsic()->setState(_calibration);
  }

  FactorBasePtr ProcessorImu2d::emplaceFactor(FeatureBasePtr _feature_motion, CaptureBasePtr _capture_origin)
  {
    CaptureImuPtr cap_imu = std::static_pointer_cast<CaptureImu>(_capture_origin);
    FeatureImu2dPtr ftr_imu = std::static_pointer_cast<FeatureImu2d>(_feature_motion);

    auto fac_imu = FactorBase::emplace<FactorImu2d>(_feature_motion, ftr_imu, cap_imu, shared_from_this(), params_->apply_loss_function);

    return fac_imu;
  }

  void ProcessorImu2d::computeCurrentDelta(const Eigen::VectorXd& _data,
                                          const Eigen::MatrixXd& _data_cov,
                                          const Eigen::VectorXd& _calib,
                                          const double _dt,
                                          Eigen::VectorXd& _delta,
                                          Eigen::MatrixXd& _delta_cov,
                                          Eigen::MatrixXd& _jac_delta_calib) const
  {
    using namespace Eigen;
    assert(_data.size() == data_size_ && "Wrong data size!");
    Vector3d data_2d;
    data_2d << _data(0), _data(1), _data(5);
    Matrix3d data_cov_2d;
    data_cov_2d << _data_cov(0,0), _data_cov(0,1), _data_cov(0,5),
                  _data_cov(1,0), _data_cov(1,1), _data_cov(1,5),
                  _data_cov(5,0), _data_cov(5,1), _data_cov(5,5);



    Matrix<double, 5, 3> jac_delta_data;
    /*
     * We have the following computing pipeline:
     *     Input: data, calib, dt
     *     Output: delta, delta_cov, jac_calib
     *
     * We do:
     *     body         = data - calib (measurement - bias)
     *     delta        = body2delta(body, dt) --> jac_body
     *     jac_data     = jac_body
     *     jac_calib    = - jac_body
     *     delta_cov  <-- propagate data_cov using jac_data
     *
     * where body2delta(.) produces a delta from body=(a,w) as follows:
     *     dp = P * a * dt^2
     *     dth = exp(w * dt) = w * dt
     *     dv = Q * a * dt
     * where P and Q are defined in imu2d::exp(), and also a jacobian J^delta_data
     */

    //create delta
    imu2d::body2delta(data_2d - _calib, _dt, _delta, jac_delta_data);

    // compute delta_cov
    _delta_cov = jac_delta_data * data_cov_2d * jac_delta_data.transpose();

    // compute jacobian_calib
    _jac_delta_calib = - jac_delta_data;
  }

  void ProcessorImu2d::deltaPlusDelta(const Eigen::VectorXd& _delta_preint,
                                      const Eigen::VectorXd& _delta,
                                      const double _dt,
                                      Eigen::VectorXd& _delta_preint_plus_delta) const
  {
    /* MATHS:
     * Simple composition,
     * (D o d)p   =   Dp + Dth*dp + Dv*dt
     * (D o d)th  =   Dth+dth
     * (D o d)v   =   Dv + Dth*dv
     * where Dth*dp and Dth*dv are rotations and not scalar products. The version with jacobians is not used here.
     */
    _delta_preint_plus_delta = imu2d::compose(_delta_preint, _delta, _dt);
  }

  void ProcessorImu2d::statePlusDelta(const VectorComposite& _x,
                                      const Eigen::VectorXd& _delta,
                                      const double _dt,
                                      VectorComposite& _x_plus_delta) const
  {
    assert(_x.includesStructure("POV") && "Any key missing in _x");
    assert(_delta.size() == 5 && "Wrong _delta vector size");
    assert(_dt >= 0 && "Time interval _dt is negative!");

    const auto& delta = VectorComposite(_delta, "POV", {2,1,2});
    /*
     * MATHS:
     *
     * In the absence of gravity (2D case) it's the same as deltaPlusDelta(), so the same compose() function is used
     */
    _x_plus_delta = imu2d::compose(_x, delta, _dt);

  }

  void ProcessorImu2d::deltaPlusDelta(const Eigen::VectorXd& _delta_preint,
                                      const Eigen::VectorXd& _delta,
                                      const double _dt,
                                      Eigen::VectorXd& _delta_preint_plus_delta,
                                      Eigen::MatrixXd& _jacobian_delta_preint,
                                      Eigen::MatrixXd& _jacobian_delta) const
  {
    /*
     * Expression of the delta integration step, D' = D (+) d:
     *
     *     Dp' = Dp + Dv*dt + Dq*dp
     *     Dv' = Dv + Dq*dv
     *     Dq' = Dq * dq
     *
     * Jacobians for covariance propagation.
     *
     * a. With respect to Delta, gives _jacobian_delta_preint = D'_D as:
     *
     *   D'_D = [ I    DR*skew(1)*dp    I*dt   
     *            0    1                0
     *            0    DR*skew(1)*dv    I   ] 
     *
     * b. With respect to delta, gives _jacobian_delta = D'_d as:
     *
     *   D'_d = [ DR   0    0
     *            0    1    0
     *            0    0    DR ]
     *
     * Note: covariance propagation, i.e.,  P+ = D_D * P * D_D' + D_d * M * D_d', is done in ProcessorMotion.
     */
    imu2d::compose(_delta_preint, _delta, _dt, _delta_preint_plus_delta, _jacobian_delta_preint, _jacobian_delta); // jacobians tested in imu2d_tools
  }

  Eigen::VectorXd ProcessorImu2d::correctDelta (const Eigen::VectorXd& delta_preint,
                                                const Eigen::VectorXd& delta_step) const
  {
    return imu2d::plus(delta_preint, delta_step);
  }


} // namespace wolf

// Register in the FactorySensor
#include "core/processor/factory_processor.h"

namespace wolf
{
  WOLF_REGISTER_PROCESSOR(ProcessorImu2d)
  WOLF_REGISTER_PROCESSOR_AUTO(ProcessorImu2d)
}
