//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include "imu/sensor/sensor_compass.h"
#include "core/state_block/state_block.h"
#include "core/state_block/state_quaternion.h"

namespace wolf
{

SensorCompass::SensorCompass(const Eigen::VectorXd& _extrinsics,
                             ParamsSensorCompassPtr _params) :
        SensorBase("SensorCompass",
                   std::make_shared<StateBlock>(_extrinsics.head(3), true),
                   std::make_shared<StateQuaternion>(_extrinsics.tail(4), true),
                   std::make_shared<StateBlock>(_params->bias_prior_state,
                                                _params->bias_prior_mode == "fix"),
                   3,
                   false, // P static
                   false, // O static
                   false),// I static
        params_compass_(_params)
{
    // Bias prior
    if (params_compass_->bias_prior_mode == "factor")
        addPriorIntrinsics(params_compass_->bias_prior_state,
                           (params_compass_->bias_prior_sigma.array() * params_compass_->bias_prior_sigma.array()).matrix().asDiagonal());

    // Global Magnetic Field 'F' state block
    addStateBlock('F',
                  std::make_shared<StateBlock>(params_compass_->field_prior_state,
                                               params_compass_->field_prior_mode == "fix"));

    if (params_compass_->field_prior_mode == "factor")
        addPriorParameter('F',
                          params_compass_->field_prior_state,
                          params_compass_->field_prior_sigma.cwiseAbs2().asDiagonal());
}

SensorCompass::SensorCompass(const Eigen::VectorXd& _extrinsics, const ParamsSensorCompass& _params) :
        SensorCompass(_extrinsics, std::make_shared<ParamsSensorCompass>(_params))
{

}

SensorCompass::~SensorCompass()
{
    // Auto-generated destructor stub
}

Eigen::Matrix3d SensorCompass::computeMeasurementCovariance(const Eigen::Vector3d& _field) const
{
    Eigen::Vector3d v1(_field.normalized());

    // Find auxiliar vector not "too parallel" to v1
    Eigen::Vector3d aux(Eigen::Vector3d::Random().normalized());
    while (v1.dot(aux) > 0.9)
        aux = Eigen::Vector3d::Random().normalized();

    // Find v2 and v3 orthogonals to v1
    Eigen::Vector3d v2, v3;
    v2 = v1.cross(aux).normalized();
    v3 = v1.cross(v2).normalized();

    // Take v1, v2 and v3 as covariance eigenvectors
    Eigen::Matrix3d eigen_vectors;
    eigen_vectors.col(0) = v1;
    eigen_vectors.col(1) = v2;
    eigen_vectors.col(2) = v3;

    // Set the covariance eigenvalues
    Eigen::Matrix3d eigen_values(Eigen::Matrix3d::Zero());
    eigen_values(0,0) = 1e6; // High covariance along field measurement direction to avoid gain problems
    eigen_values(1,1) = pow(params_compass_->stdev_noise,2);
    eigen_values(2,2) = pow(params_compass_->stdev_noise,2);

    return eigen_vectors * eigen_values * eigen_vectors.inverse();
}

} /* namespace wolf */

// Register in the FactorySensor
#include "core/sensor/factory_sensor.h"
namespace wolf {
WOLF_REGISTER_SENSOR(SensorCompass);
WOLF_REGISTER_SENSOR_AUTO(SensorCompass);
} // namespace wolf

