//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include "core/utils/utils_gtest.h"

#include "imu/internal/config.h"
#include "imu/factor/factor_compass_3d.h"
#include "imu/capture/capture_compass.h"
#include "imu/sensor/sensor_compass.h"
#include "imu/processor/processor_compass.h"

#include "core/ceres_wrapper/solver_ceres.h"
#include "core/math/rotations.h"

using namespace Eigen;
using namespace wolf;
using std::cout;
using std::endl;

std::string wolf_root = _WOLF_IMU_ROOT_DIR;

// Problem and solver
ProblemPtr problem_ptr = Problem::create("PO", 3);
SolverCeres solver(problem_ptr);

// Sensor
auto sen = std::static_pointer_cast<SensorCompass>(problem_ptr->installSensor("SensorCompass",
                                                                              "magnetometer",
                                                                              (Vector7d() << 0,0,0,0,0,0,1).finished(),
                                                                              std::make_shared<ParamsSensorCompass>()));

// Frame
FrameBasePtr frm = problem_ptr->emplaceFrame(0, (Vector7d() << 0,0,0,0,0,0,1).finished() );

// Pointers
CaptureCompassPtr cap;

// Auxliary states
Vector3d frm_p, sen_p, sen_bias, field, meas;
Quaterniond frm_q, sen_q;

void generateRandomSetup()
{
    // remove capture (and feature and factor) just in case
    if (cap)
        cap->remove();

    // Random states
    frm_p = Vector3d::Random() * 10;
    frm_q = Quaterniond::UnitRandom();
    sen_p = Vector3d::Random() * 10;
    sen_q = Quaterniond::UnitRandom();
    sen_bias = Vector3d::Random();
    field = Vector3d::Random().normalized() * 50;

    // measurement
    meas = sen_q.conjugate() * frm_q.conjugate() * field + sen_bias;

    // Set states
    frm->getP()->setState(frm_p);
    frm->getO()->setState(frm_q.coeffs());
    sen->getP()->setState(sen_p);
    sen->getO()->setState(sen_q.coeffs());
    sen->getIntrinsic()->setState(sen_bias);
    sen->getStateBlock('F')->setState(field);

    WOLF_DEBUG("Random setup:",
               "\n\tfrm: P = ", frm_p.transpose(), " | O = ", frm_q.coeffs().transpose(),
               "\n\tsen: P = ", sen_p.transpose(), " | O = ", sen_q.coeffs().transpose(), " | I = ", sen_bias.transpose(),
               "\n\tfield: F = ", field.transpose(),
               "\n\tmeasurement = ", meas.transpose());

    // Capture
    cap = CaptureBase::emplace<CaptureCompass>(frm,
                                               1,
                                               sen,
                                               meas,
                                               Matrix3d::Identity());

    // feature & factor with delta measurement
    auto fea = FeatureBase::emplace<FeatureBase>(cap,
                                                 "FeatureCompass",
                                                 meas,
                                                 Eigen::Matrix3d::Identity());

    FactorBase::emplace<FactorCompass3d>(fea,
                                         fea,
                                         nullptr,
                                         false);

    // Fix all
    frm->getP()->fix();
    frm->getO()->fix();
    sen->getP()->fix();
    sen->getO()->fix();
    sen->getIntrinsic()->fix();
    sen->getStateBlock('F')->fix();
}

TEST(FactorCompass3d, solve_frm_q)
{
    for (int i = 0; i < 1e2; i++)
    {
        // Random setup
        generateRandomSetup();

        // solve
        std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

        //WOLF_INFO(report);

        ASSERT_MATRIX_APPROX(frm->getO()->getState(), frm_q.coeffs(), 1e-6);

        // Unfix and perturb: frame orientation
        frm->getO()->unfix();
        frm->getO()->perturb();

        // solve
        report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

        //WOLF_INFO(report);

        // check measurement and field are aligned
        Vector3d exp_meas = sen_q.conjugate() * frm_q.conjugate() * field + sen_bias;

        ASSERT_MATRIX_APPROX(meas, exp_meas, 1e-6);

        // remove capture (and feature and factor) for the next loop
        cap->remove();
    }
}


TEST(FactorCompass3d, solve_sen_q)
{
    for (int i = 0; i < 1e2; i++)
    {
        // Random setup
        generateRandomSetup();

        // solve
        std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

        //WOLF_INFO(report);

        ASSERT_MATRIX_APPROX(frm->getO()->getState(), frm_q.coeffs(), 1e-6);

        // Unfix and perturb: frame orientation
        sen->getO()->unfix();
        sen->getO()->perturb();

        // solve
        report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

        //WOLF_INFO(report);

        // check measurement and field are aligned
        Vector3d exp_meas = sen_q.conjugate() * frm_q.conjugate() * field + sen_bias;

        ASSERT_MATRIX_APPROX(meas, exp_meas, 1e-6);

        // remove capture (and feature and factor) for the next loop
        cap->remove();
    }
}

TEST(FactorCompass3d, solve_bias)
{
    for (int i = 0; i < 1e2; i++)
    {
        // Random setup
        generateRandomSetup();

        // Unfix and perturb: bias
        sen->getIntrinsic()->unfix();
        sen->getIntrinsic()->perturb(1);

        // solve
        std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

        //WOLF_INFO(report);

        ASSERT_MATRIX_APPROX(sen->getIntrinsic()->getState(), sen_bias, 1e-6);

        // remove capture (and feature and factor) for the next loop
        cap->remove();
    }
}

TEST(FactorCompass3d, solve_field)
{
    for (int i = 0; i < 1e2; i++)
    {
        // Random setup
        generateRandomSetup();

        // Unfix and perturb: bias
        sen->getStateBlock('F')->unfix();
        sen->getStateBlock('F')->perturb(10);

        // solve
        std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

        //WOLF_INFO(report);

        ASSERT_MATRIX_APPROX(sen->getStateBlock('F')->getState(), field, 1e-6);

        // remove capture (and feature and factor) for the next loop
        cap->remove();
    }
}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
