//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include <core/ceres_wrapper/solver_ceres.h>
#include <core/utils/utils_gtest.h>

#include "imu/factor/factor_imu2d.h"
#include "imu/math/imu2d_tools.h"
#include "imu/sensor/sensor_imu2d.h"

using namespace Eigen;
using namespace wolf;

// Input odometry data and covariance
Matrix6d data_cov = 0.1*Matrix6d::Identity();
Matrix5d delta_cov = 0.1*Matrix5d::Identity();

// Jacobian of the bias
MatrixXd jacobian_bias((MatrixXd(5,3) << 1, 0, 0,
                                         0, 1, 0,
                                         0, 0, 1,
                                         1, 0, 0,
                                         0, 1, 0 ).finished());
//preintegration bias
Vector3d b0_preint = Vector3d::Zero();

// Problem and solver
ProblemPtr problem_ptr = Problem::create("POV", 2);
SolverCeres solver(problem_ptr);

// Two frames
FrameBasePtr frm0 = problem_ptr->emplaceFrame(TimeStamp(0), Vector5d::Zero());
FrameBasePtr frm1 = problem_ptr->emplaceFrame(TimeStamp(1), Vector5d::Zero());

// Imu2d sensor
SensorBasePtr sensor = std::make_shared<SensorImu2d>( Vector3d::Zero(), ParamsSensorImu2d() ); // default params: see sensor_imu2d.h

//capture from frm1 to frm0
auto cap0 = CaptureBase::emplace<CaptureImu>(frm0, 0, sensor, Vector6d::Zero(), data_cov, Vector3d::Zero(), nullptr);
auto cap1 = CaptureBase::emplace<CaptureImu>(frm1, 1, sensor, Vector6d::Zero(), data_cov, Vector3d::Zero(), cap0);

TEST(FactorImu2d, check_tree)
{
    ASSERT_TRUE(problem_ptr->check(0));
}

TEST(FactorImu2d, bias_zero_solve_f1)
{
  for(int i = 0; i < 50; i++){
    // Random delta
    Vector5d delta = Vector5d::Random() * 10; //delta *= 0;
    delta(2) = pi2pi(delta(2));

    // Random initial pose
    Vector5d x0 = Vector5d::Random() * 10; //x0 *= 0;
    x0(2) = pi2pi(x0(2));

    // x1 groundtruth
    Vector5d x1;
    x1 = imu2d::compose(x0, delta, 1);

    // Set states
    frm0->setState(x0);
    frm1->setState(x1);

    // feature & factor with delta measurement
    auto fea1 = FeatureBase::emplace<FeatureImu2d>(cap1, delta, delta_cov, b0_preint, jacobian_bias, cap1);
    FactorBase::emplace<FactorImu2d>(fea1, fea1, cap0, nullptr, false);

    // Fix frm0 and biases, perturb frm1
    frm0->fix();
    cap0->fix();
    frm1->unfix();
    cap1->fix();
    frm1->perturb(0.01);

    // solve 
    std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

    ASSERT_POSE2d_APPROX(frm1->getStateVector(), x1, 1e-6);
    //WOLF_INFO(frm1->getStateVector());

    // remove feature (and factor) for the next loop
    fea1->remove();
}
}

TEST(FactorImu2d, bias_zero_solve_f0)
{
  for(int i = 0; i < 50; i++){
    // Random delta
    Vector5d delta = Vector5d::Random() * 10; //delta *= 0;
    delta(2) = pi2pi(delta(2));

    // Random initial pose
    Vector5d x0 = Vector5d::Random() * 10; //x0 *= 0;
    x0(2) = pi2pi(x0(2));

    // x1 groundtruth
    Vector5d x1;
    x1 = imu2d::compose(x0, delta, 1);

    // Set states
    frm0->setState(x0);
    frm1->setState(x1);

    // feature & factor with delta measurement
    auto fea1 = FeatureBase::emplace<FeatureImu2d>(cap1, delta, delta_cov, b0_preint, jacobian_bias, cap1);
    FactorBase::emplace<FactorImu2d>(fea1, fea1, cap0, nullptr, false);

    // Fix frm0 and biases, perturb frm1
    frm0->unfix();
    cap0->fix();
    frm1->fix();
    cap1->fix();
    frm0->perturb(0.01);

    // solve 
    std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

    ASSERT_POSE2d_APPROX(frm0->getStateVector(), x0, 1e-6);
    //WOLF_INFO(frm1->getStateVector());

    // remove feature (and factor) for the next loop
    fea1->remove();
}
}

TEST(FactorImu2d, bias_zero_solve_b1)
{
  for(int i = 0; i < 50; i++){
    // Random delta
    Vector5d delta = Vector5d::Random() * 10; //delta *= 0;
    delta(2) = pi2pi(delta(2));

    // Random initial pose
    Vector5d x0 = Vector5d::Random() * 10; //x0 *= 0;
    x0(2) = pi2pi(x0(2));

    // x1 groundtruth
    Vector5d x1;
    x1 = imu2d::compose(x0, delta, 1);
    
    //0 Initial bias
    Vector3d b0 = Vector3d::Zero();

    // Set states
    frm0->setState(x0);
    frm1->setState(x1);
    cap0->getStateBlock('I')->setState(b0);
    cap1->getStateBlock('I')->setState(b0);

    // feature & factor with delta measurement
    auto fea1 = FeatureBase::emplace<FeatureImu2d>(cap1, delta, delta_cov, b0_preint, jacobian_bias, cap1);
    FactorBase::emplace<FactorImu2d>(fea1, fea1, cap0, nullptr, false);

    // Fix frm0, frm1 and cap0, unfix cap1 (bias), perturb cap1
    frm0->fix();
    cap0->fix();
    frm1->fix();
    cap1->unfix();
    cap1->perturb(0.01);

    // solve  to estimate bias at cap1
    std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

    ASSERT_MATRIX_APPROX(cap1->getStateVector(), b0, 1e-6);
    //WOLF_INFO(cap1->getStateVector());

    // remove feature (and factor) for the next loop
    fea1->remove();
}
}

TEST(FactorImu2d, solve_b0)
{
  for(int i = 0; i < 50; i++){
    // Random delta
    Vector5d delta_biased = Vector5d::Random() * 10;
    delta_biased(2) = pi2pi(delta_biased(2));

    // Random initial pose
    Vector5d x0 = Vector5d::Random() * 10;
    x0(2) = pi2pi(x0(2));

    //Random Initial bias
    Vector3d b0 = Vector3d::Random();

    //Corrected delta
    Vector5d delta_step = jacobian_bias*(b0-b0_preint);
    Vector5d delta = imu2d::plus(delta_biased, delta_step);

    // x1 groundtruth
    Vector5d x1;
    x1 = imu2d::compose(x0, delta, 1);

    // Set states
    frm0->setState(x0);
    frm1->setState(x1);
    cap0->getStateBlock('I')->setState(b0);
    cap1->getStateBlock('I')->setState(b0);

    // feature & factor with delta measurement
    auto fea1 = FeatureBase::emplace<FeatureImu2d>(cap1, delta_biased, delta_cov, b0_preint, jacobian_bias);
    FactorBase::emplace<FactorImu2d>(fea1, fea1, cap0, nullptr, false);

    // Fix frm1, perturb frm0
    frm0->fix();
    cap0->unfix();
    frm1->fix();
    cap1->fix();
    cap0->perturb(0.1);

    // solve 
    std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

    ASSERT_POSE2d_APPROX(cap0->getStateVector(), b0, 1e-6);
    //WOLF_INFO(cap0->getStateVector());

    // remove feature (and factor) for the next loop
    fea1->remove();
}
}

TEST(FactorImu2d, solve_b1)
{
  for(int i = 0; i < 50; i++){
    // Random delta
    Vector5d delta = Vector5d::Random() * 10; //delta *= 0;
    delta(2) = pi2pi(delta(2));

    // Random initial pose
    Vector5d x0 = Vector5d::Random() * 10; //x0 *= 0;
    x0(2) = pi2pi(x0(2));

    // x1 groundtruth
    Vector5d x1;
    x1 = imu2d::compose(x0, delta, 1);
    
    //0 Initial bias
    Vector3d b0 = Vector3d::Random();

    // Set states
    frm0->setState(x0);
    frm1->setState(x1);
    cap0->getStateBlock('I')->setState(b0);
    cap1->getStateBlock('I')->setState(b0);

    // feature & factor with delta measurement
    auto fea1 = FeatureBase::emplace<FeatureImu2d>(cap1, delta, delta_cov, b0_preint, jacobian_bias, cap1);
    FactorBase::emplace<FactorImu2d>(fea1, fea1, cap0, nullptr, false);

    // Fix frm0 and frm1, unfix bias, perturb cap1
    frm0->fix();
    cap0->fix();
    frm1->fix();
    cap1->unfix();
    cap1->perturb(0.01);

    // solve  to estimate bias at cap1
    std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

    ASSERT_MATRIX_APPROX(cap1->getStateVector(), b0, 1e-6);
    //WOLF_INFO(cap1->getStateVector());

    // remove feature (and factor) for the next loop
    fea1->remove();
}
}

TEST(FactorImu2d, solve_f0)
{
  for(int i = 0; i < 50; i++){
    // Random delta
    Vector5d delta_biased = Vector5d::Random() * 10;
    delta_biased(2) = pi2pi(delta_biased(2));

    // Random initial pose
    Vector5d x0 = Vector5d::Random() * 10;
    x0(2) = pi2pi(x0(2));

    //Random Initial bias
    Vector3d b0 = Vector3d::Random();

    //Corrected delta
    Vector5d delta_step = jacobian_bias*(b0-b0_preint);
    Vector5d delta = imu2d::plus(delta_biased, delta_step);

    // x1 groundtruth
    Vector5d x1;
    x1 = imu2d::compose(x0, delta, 1);

    // Set states
    frm0->setState(x0);
    frm1->setState(x1);
    cap0->getStateBlock('I')->setState(b0);
    cap1->getStateBlock('I')->setState(b0);

    // feature & factor with delta measurement
    auto fea1 = FeatureBase::emplace<FeatureImu2d>(cap1, delta_biased, delta_cov, b0_preint, jacobian_bias);
    FactorBase::emplace<FactorImu2d>(fea1, fea1, cap0, nullptr, false);

    // Fix frm1, perturb frm0
    frm0->unfix();
    cap0->fix();
    frm1->fix();
    cap1->fix();
    frm0->perturb(0.1);

    // solve 
    std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

    ASSERT_POSE2d_APPROX(frm0->getStateVector(), x0, 1e-6);
    //WOLF_INFO(frm0->getStateVector());

    // remove feature (and factor) for the next loop
    fea1->remove();
}
}

TEST(FactorImu2d, solve_f1)
{
  for(int i = 0; i < 50; i++){
    // Random delta
    Vector5d delta_biased = Vector5d::Random() * 10;
    delta_biased(2) = pi2pi(delta_biased(2));

    // Random initial pose
    Vector5d x0 = Vector5d::Random() * 10;
    x0(2) = pi2pi(x0(2));

    //Random Initial bias
    Vector3d b0 = Vector3d::Random();

    //Corrected delta
    Vector5d delta_step = jacobian_bias*(b0-b0_preint);
    Vector5d delta = imu2d::plus(delta_biased, delta_step);

    // x1 groundtruth
    Vector5d x1;
    x1 = imu2d::compose(x0, delta, 1);

    // Set states
    frm0->setState(x0);
    frm1->setState(x1);
    cap0->getStateBlock('I')->setState(b0);
    cap1->getStateBlock('I')->setState(b0);

    // feature & factor with delta measurement
    auto fea1 = FeatureBase::emplace<FeatureImu2d>(cap1, delta_biased, delta_cov, b0_preint, jacobian_bias);
    FactorBase::emplace<FactorImu2d>(fea1, fea1, cap0, nullptr, false);

    // Fix frm1, perturb frm0
    frm0->fix();
    cap0->fix();
    frm1->unfix();
    cap1->fix();
    frm1->perturb(0.1);

    // solve 
    std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

    ASSERT_POSE2d_APPROX(frm1->getStateVector(), x1, 1e-6);
    //WOLF_INFO(frm0->getStateVector());

    // remove feature (and factor) for the next loop
    fea1->remove();
}
}

TEST(FactorImu2d, solve_f1_b1)
{
  for(int i = 0; i < 50; i++){
    // Random delta
    Vector5d delta_biased = Vector5d::Random() * 10;
    delta_biased(2) = pi2pi(delta_biased(2));

    // Random initial pose
    Vector5d x0 = Vector5d::Random() * 10;
    x0(2) = pi2pi(x0(2));

    //Random Initial bias
    Vector3d b0 = Vector3d::Random();

    //Corrected delta
    Vector5d delta_step = jacobian_bias*(b0-b0_preint);
    Vector5d delta = imu2d::plus(delta_biased, delta_step);

    // x1 groundtruth
    Vector5d x1;
    x1 = imu2d::compose(x0, delta, 1);

    // Set states
    frm0->setState(x0);
    frm1->setState(x1);
    cap0->getStateBlock('I')->setState(b0);
    cap1->getStateBlock('I')->setState(b0);

    // feature & factor with delta measurement
    auto fea1 = FeatureBase::emplace<FeatureImu2d>(cap1, delta_biased, delta_cov, b0_preint, jacobian_bias);
    FactorBase::emplace<FactorImu2d>(fea1, fea1, cap0, nullptr, false);

    // Fix frm1, perturb frm0
    frm0->fix();
    cap0->fix();
    frm1->unfix();
    cap1->unfix();
    frm1->perturb(0.1);
    cap1->perturb(0.1);

    // solve 
    std::string report = solver.solve(SolverManager::ReportVerbosity::BRIEF);

    ASSERT_POSE2d_APPROX(frm1->getStateVector(), x1, 1e-6);
    ASSERT_MATRIX_APPROX(cap1->getStateVector(), b0, 1e-6);
    //WOLF_INFO(frm0->getStateVector());

    // remove feature (and factor) for the next loop
    fea1->remove();
  }
}

int main(int argc, char **argv)
{
    testing::InitGoogleTest(&argc, argv);
    //    ::testing::GTEST_FLAG(filter) = "FactorImu2d.no_bias_fixed"; // Test only this one
    return RUN_ALL_TESTS();
}
