//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/*
 * gtest_imu2d_static_init.cpp
 *
 *  Created on: Sept 2021
 *      Author: igeer
 */

#include <fstream>
#include <core/ceres_wrapper/solver_ceres.h>
#include <core/utils/utils_gtest.h>
#include "imu/internal/config.h"

#include "imu/factor/factor_imu2d.h"
#include "imu/math/imu2d_tools.h"
#include "imu/sensor/sensor_imu2d.h"
#include "core/capture/capture_void.h"
#include <core/factor/factor_relative_pose_2d.h>

using namespace Eigen;
using namespace wolf;

class ProcessorImu2dStaticInitTest : public testing::Test
{

    public: //These can be accessed in fixtures
        wolf::ProblemPtr problem_ptr_;
        wolf::SensorBasePtr sensor_ptr_;
        wolf::ProcessorMotionPtr processor_motion_;
        wolf::TimeStamp t;
        wolf::TimeStamp t0;
        double dt;
        Vector6d data;
        Matrix6d data_cov;
        FrameBasePtr KF0_;
        FrameBasePtr first_frame_;
        FrameBasePtr last_frame_;
        SolverCeresPtr solver_;

    //a new of this will be created for each new test
    void SetUp() override
    {
        WOLF_INFO("Doing setup...");
        using namespace Eigen;
        using std::shared_ptr;
        using std::make_shared;
        using std::static_pointer_cast;
        using namespace wolf::Constants;

        std::string wolf_root = _WOLF_IMU_ROOT_DIR;

        // Wolf problem
        problem_ptr_ = Problem::create("POV", 2);
        Vector3d extrinsics = (Vector3d() << 0,0,0).finished();
        sensor_ptr_ = problem_ptr_->installSensor("SensorImu2d", "Main Imu", extrinsics,  wolf_root + "/test/yaml/sensor_imu2d_static_init.yaml");
        ProcessorBasePtr processor_ptr = problem_ptr_->installProcessor("ProcessorImu2d", "Imu pre-integrator", "Main Imu", wolf_root + "/test/yaml/imu2d_static_init.yaml");
        processor_motion_ = std::static_pointer_cast<ProcessorMotion>(processor_ptr);

        // Time and data variables
        dt = 0.1;
        t0.set(0);
        t = t0;
        data = Vector6d::Random();
        data_cov = Matrix6d::Identity();
        last_frame_ = nullptr;
        first_frame_ = nullptr;

        // Create the solver
        solver_ = make_shared<SolverCeres>(problem_ptr_);
        
        // Set the origin
        VectorComposite x_origin("POV", {Vector2d::Zero(), Vector1d::Zero(), Vector2d::Zero()});
        VectorComposite std_origin("POV", {0.001*Vector2d::Ones(), 0.001*Vector1d::Ones(), 0.001*Vector2d::Ones()});
        //KF0_ = problem_ptr_->setPriorFix(x_origin, t0);
        KF0_ = problem_ptr_->setPriorFactor(x_origin, std_origin, 0);

    }

    void TestStatic(const Vector3d& body_magnitudes, const Vector3d& bias_groundtruth, const Vector3d& bias_initial_guess, const string& test_name, int testing_var, bool small_tol)
    {
      //Data
      data.head(2) = body_magnitudes.head(2);
      data(5) = body_magnitudes(2);
      data.head(2) += bias_groundtruth.head(2);
      data(5) += bias_groundtruth(2);

      //Set bias initial guess
      sensor_ptr_->getIntrinsic(t0)->setState(bias_initial_guess);
      processor_motion_->setOrigin(KF0_);

#if WRITE_CSV_FILE
      std::fstream file_est; 
      file_est.open("./est2d-"+test_name+".csv", std::fstream::out);
      //    std::string header_est = "t,px,py,pz,qx,qy,qz,qw,vx,vy,vz,bax_est,bax_preint\n";
      std::string header_est;
      if(testing_var == 0) header_est = "t;px;vx;o;bax_est;bg_est;bax_preint;bg_preint\n";
      if(testing_var == 1) header_est = "t;py;vy;o;bay_est;bg_est;bay_preint;bg_preint\n";
      //if(testing_var == 2) header_est = "t;pz;vz;oz:baz_est;bgz_est;baz_preint;bgz_preint\n";
      if(testing_var == 3) header_est = "t;pnorm;vnorm;o;banorm_est;bg_est;banorm_preint;bg_preint\n";
      file_est << header_est;
#endif


      int n_frames = 0;
      for(t = t0; t <= t0 + 9.9 + dt/2; t+=dt){
        WOLF_INFO("\n------------------------------------------------------------------------");

        //Create and process capture  
        auto C = std::make_shared<CaptureImu>(t, sensor_ptr_, data, data_cov, KF0_->getCaptureList().front());
        C->process();

        auto state = problem_ptr_->getState();
        auto bias_est = sensor_ptr_->getIntrinsic()->getState();
        auto bias_preint = processor_motion_->getLast()->getCalibrationPreint();

#if WRITE_CSV_FILE
        // pre-solve print to CSV
        if(testing_var == 3){
        file_est << std::fixed << t << ";"
            << state['P'].norm() << ";"
            << state['V'].norm() << ";"
            << state['O'] << ";"
            << bias_est.head(2).norm() << ";"
            << bias_est(2) << ";"
            << bias_preint.head(2).norm() << ";"
            << bias_preint(2) << "\n";
        }
        else
        {
        file_est << std::fixed << t << ";"
            << state['P'](testing_var) << ";"
            << state['V'](testing_var) << ";"
            << state['O'] << ";"
            << bias_est(testing_var) << ";"
            << bias_est(2) << ";"
            << bias_preint(testing_var) << ";"
            << bias_preint(2) << "\n";

        }
#endif

        // new frame
        if (last_frame_ != processor_motion_->getOrigin()->getFrame())
        {
          n_frames++;
          last_frame_ = processor_motion_->getOrigin()->getFrame();

          // impose static
          last_frame_->getP()->setState(KF0_->getP()->getState());
          last_frame_->getO()->setState(KF0_->getO()->getState());
          last_frame_->getV()->setZero();

          //Fix frame
          last_frame_->fix();

          //Solve
          if(n_frames > 0)
          {
            std::string report = solver_->solve(SolverManager::ReportVerbosity::FULL);
            //WOLF_INFO("Solver Report: ", report);

            state = problem_ptr_->getState();
            bias_est = sensor_ptr_->getIntrinsic()->getState();
            bias_preint = processor_motion_->getLast()->getCalibrationPreint();
            
            WOLF_INFO("The current problem is:");
            problem_ptr_->print(4);

#if WRITE_CSV_FILE
            // post-solve print to CSV with time-stamp shifted by dt/2 to separate from pre-solve  result
            if(testing_var == 3){
              file_est << std::fixed << t+dt/2 << ";"
                << state['P'].norm() << ";"
                << state['V'].norm() << ";"
                << state['O'] << ";"
                << bias_est.head(2).norm() << ";"
                << bias_est(2) << ";"
                << bias_preint.head(2).norm() << ";"
                << bias_preint(2) << "\n";
            }
            else
            {
              file_est << std::fixed << t+dt/2 << ";"
                << state['P'](testing_var) << ";"
                << state['V'](testing_var) << ";"
                << state['O'] << ";"
                << bias_est(testing_var) << ";"
                << bias_est(2) << ";"
                << bias_preint(testing_var) << ";"
                << bias_preint(2) << "\n";

            }
#endif
          }

        }


        WOLF_INFO("Number of frames is ", n_frames);
        WOLF_INFO("The state is: ", state);
        WOLF_INFO("The estimated bias is: ", bias_est.transpose());
        WOLF_INFO("The preintegrated bias is: ", bias_preint.transpose());
        if(small_tol)
        {
          if(n_frames == 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-6);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-6);
          }
          else if (n_frames > 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-6);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-6);
            EXPECT_MATRIX_APPROX(bias_preint, bias_groundtruth, 1e-6);
          }
        }
        else
        {
          if(n_frames == 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-3);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-3);
          }
          else if (n_frames > 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-3);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-3);
            EXPECT_MATRIX_APPROX(bias_preint, bias_groundtruth, 1e-3);
          }
        }
      }

#if WRITE_CSV_FILE
      file_est.close();
#endif

    }

    void TestStaticZeroOdom(const Vector3d& body_magnitudes, const Vector3d& bias_groundtruth, const Vector3d& bias_initial_guess, const string& test_name, int testing_var, bool small_tol)
    {
      //Data
      data.head(2) = body_magnitudes.head(2);
      data(5) = body_magnitudes(2);
      data.head(2) += bias_groundtruth.head(2);
      data(5) += bias_groundtruth(2);

      //Set bias initial guess
      sensor_ptr_->getIntrinsic(t0)->setState(bias_initial_guess);
      processor_motion_->setOrigin(KF0_);

#if WRITE_CSV_FILE
      std::fstream file_est; 
      file_est.open("./est2dzeroodom-"+test_name+".csv", std::fstream::out);
      //    std::string header_est = "t,px,py,pz,qx,qy,qz,qw,vx,vy,vz,bax_est,bax_preint\n";
      std::string header_est;
      if(testing_var == 0) header_est = "t;px;vx;o;bax_est;bg_est;bax_preint;bg_preint\n";
      if(testing_var == 1) header_est = "t;py;vy;o;bay_est;bg_est;bay_preint;bg_preint\n";
      //if(testing_var == 2) header_est = "t;pz;vz;oz:baz_est;bgz_est;baz_preint;bgz_preint\n";
      if(testing_var == 3) header_est = "t;pnorm;vnorm;o;banorm_est;bg_est;banorm_preint;bg_preint\n";
      file_est << header_est;
#endif

      int n_frames = 0;
      for(t = t0; t <= t0 + 9.9 + dt/2; t+=dt){
        WOLF_INFO("\n------------------------------------------------------------------------");

        //Create and process capture  
        auto C = std::make_shared<CaptureImu>(t, sensor_ptr_, data, data_cov, KF0_->getCaptureList().front());
        C->process();

        auto state = problem_ptr_->getState();
        auto bias_est = sensor_ptr_->getIntrinsic()->getState();
        auto bias_preint = processor_motion_->getLast()->getCalibrationPreint();

#if WRITE_CSV_FILE
        // pre-solve print to CSV
        if(testing_var == 3){
        file_est << std::fixed << t << ";"
            << state['P'].norm() << ";"
            << state['V'].norm() << ";"
            << state['O'] << ";"
            << bias_est.head(2).norm() << ";"
            << bias_est(2) << ";"
            << bias_preint.head(2).norm() << ";"
            << bias_preint(2) << "\n";
        }
        else
        {
        file_est << std::fixed << t << ";"
            << state['P'](testing_var) << ";"
            << state['V'](testing_var) << ";"
            << state['O'] << ";"
            << bias_est(testing_var) << ";"
            << bias_est(2) << ";"
            << bias_preint(testing_var) << ";"
            << bias_preint(2) << "\n";

        }
#endif

        // new frame
        if (last_frame_ != processor_motion_->getOrigin()->getFrame())
        {
          n_frames++;
          last_frame_ = processor_motion_->getOrigin()->getFrame();
          
          // impose zero odometry
          processor_motion_->setOdometry(sensor_ptr_->getProblem()->stateZero(processor_motion_->getStateStructure()));

          // add zero displacement and rotation capture & feature & factor with all previous frames
          assert(sensor_ptr_->getProblem());
          for (auto frm_pair =  sensor_ptr_->getProblem()->getTrajectory()->begin();
              frm_pair != sensor_ptr_->getProblem()->getTrajectory()->end();
              frm_pair++)
          {
            if (frm_pair->second == last_frame_)
              break;
            auto capture_zero = CaptureBase::emplace<CaptureVoid>(last_frame_, last_frame_->getTimeStamp(), nullptr);

            auto feature_zero = FeatureBase::emplace<FeatureBase>(capture_zero,
                "FeatureZeroOdom",
                Vector3d::Zero(),
                Eigen::MatrixXd::Identity(3,3) * 0.01);

            FactorBase::emplace<FactorRelativePose2d>(feature_zero,
                feature_zero,
                frm_pair->second,
                nullptr,
                false,
                TOP_MOTION);

          }

          // impose static
          last_frame_->getV()->setZero();

          //Fix frame
          last_frame_->getV()->fix();

          //Solve
          if(n_frames > 0)
          {
            std::string report = solver_->solve(SolverManager::ReportVerbosity::FULL);
            //WOLF_INFO("Solver Report: ", report);

            state = problem_ptr_->getState();
            bias_est = sensor_ptr_->getIntrinsic()->getState();
            bias_preint = processor_motion_->getLast()->getCalibrationPreint();
            
            WOLF_INFO("The current problem is:");
            problem_ptr_->print(4);

#if WRITE_CSV_FILE
            // post-solve print to CSV with time-stamp shifted by dt/2 to separate from pre-solve  result
            if(testing_var == 3){
              file_est << std::fixed << t+dt/2 << ";"
                << state['P'].norm() << ";"
                << state['V'].norm() << ";"
                << state['O'] << ";"
                << bias_est.head(2).norm() << ";"
                << bias_est(2) << ";"
                << bias_preint.head(2).norm() << ";"
                << bias_preint(2) << "\n";
            }
            else
            {
              file_est << std::fixed << t+dt/2 << ";"
                << state['P'](testing_var) << ";"
                << state['V'](testing_var) << ";"
                << state['O'] << ";"
                << bias_est(testing_var) << ";"
                << bias_est(2) << ";"
                << bias_preint(testing_var) << ";"
                << bias_preint(2) << "\n";

            }
#endif
          }

        }



        WOLF_INFO("Number of frames is ", n_frames);
        WOLF_INFO("The state is: ", state);
        WOLF_INFO("The estimated bias is: ", bias_est.transpose());
        WOLF_INFO("The preintegrated bias is: ", bias_preint.transpose());
        if(small_tol)
        {
          if(n_frames == 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-6);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-6);
          }
          else if (n_frames > 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-6);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-6);
            EXPECT_MATRIX_APPROX(bias_preint, bias_groundtruth, 1e-6);
          }
        }
        else
        {
          if(n_frames == 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-3);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-3);
          }
          else if (n_frames > 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-3);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-3);
            EXPECT_MATRIX_APPROX(bias_preint, bias_groundtruth, 1e-3);
          }
        }
      }

#if WRITE_CSV_FILE
      file_est.close();
#endif

    }
};



TEST_F(ProcessorImu2dStaticInitTest, static_bias_aX_initial_guess_zero)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = (Vector3d() << 0.42, 0, 0).finished(); 
   Vector3d bias_initial_guess = Vector3d::Zero();

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_aX_initial_guess_zero", 0, true); 
}

TEST_F(ProcessorImu2dStaticInitTest, static_bias_zero_initial_guess_aX)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = Vector3d::Zero();
   Vector3d bias_initial_guess = (Vector3d() << 0.42, 0, 0).finished(); 

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_aX", 0, true); 
}

TEST_F(ProcessorImu2dStaticInitTest, static_bias_gX_initial_guess_zero)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = (Vector3d() << 0, 0, 0.01).finished(); 
   Vector3d bias_initial_guess = Vector3d::Zero();

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_gX_initial_guess_zero", 0, false); 
}

TEST_F(ProcessorImu2dStaticInitTest, static_bias_zero_initial_guess_gX)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = Vector3d::Zero();
   Vector3d bias_initial_guess = (Vector3d() << 0, 0, 0.01).finished(); 

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_gX", 0, false); 
}

TEST_F(ProcessorImu2dStaticInitTest, static_bias_aX_initial_guess_zero_zeroodom)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = (Vector3d() << 0.42, 0, 0).finished(); 
   Vector3d bias_initial_guess = Vector3d::Zero();

   TestStaticZeroOdom(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_aX_initial_guess_zero", 0, true); 
}

TEST_F(ProcessorImu2dStaticInitTest, static_bias_zero_initial_guess_aX_zeroodom)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = Vector3d::Zero();
   Vector3d bias_initial_guess = (Vector3d() << 0.42, 0, 0).finished(); 

   TestStaticZeroOdom(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_aX", 0, true); 
}

TEST_F(ProcessorImu2dStaticInitTest, static_bias_gX_initial_guess_zero_zeroodom)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = (Vector3d() <<  0, 0, 0.01).finished(); 
   Vector3d bias_initial_guess = Vector3d::Zero();

   TestStaticZeroOdom(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_gX_initial_guess_zero", 0, false); 
}

TEST_F(ProcessorImu2dStaticInitTest, static_bias_zero_initial_guess_gX_zeroodom)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = Vector3d::Zero();
   Vector3d bias_initial_guess = (Vector3d() << 0, 0, 0.01).finished(); 

   TestStaticZeroOdom(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_gX", 0, false); 
}


TEST_F(ProcessorImu2dStaticInitTest, static_bias_zero_initial_guess_a_random)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = Vector3d::Zero();
   Vector3d bias_initial_guess = Vector3d::Random()*100;
   bias_initial_guess(2) = 0;

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_a_random", 3, true); 
}

TEST_F(ProcessorImu2dStaticInitTest, static_bias_zero_initial_guess_random)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = Vector3d::Zero();
   Vector3d bias_initial_guess = Vector3d::Random()*0.01;

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_random", 3, false); 
}

TEST_F(ProcessorImu2dStaticInitTest, realistic_test)
{
   Vector3d body_magnitudes = Vector3d::Zero();
   Vector3d bias_groundtruth = (Vector3d() << -0.529550648247,
                                               0.278316717683,
                                              -0.00122491355676).finished();
   Vector3d bias_initial_guess = Vector3d::Zero();

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_random", 3, false); 
}

int main(int argc, char **argv)
{
    testing::InitGoogleTest(&argc, argv);
    //    ::testing::GTEST_FLAG(filter) = "FactorImu2d.no_bias_fixed"; // Test only this one
    return RUN_ALL_TESTS();
}
