//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/*
 * gtest_imu_static_init.cpp
 *
 *  Created on: Sept 2021
 *      Author: igeer
 */
#include <fstream>

#include <core/ceres_wrapper/solver_ceres.h>
#include <core/utils/utils_gtest.h>
#include "imu/internal/config.h"

#include "imu/factor/factor_imu.h"
#include "imu/math/imu_tools.h"
#include "imu/sensor/sensor_imu.h"
#include "core/capture/capture_void.h"
#include <core/factor/factor_relative_pose_3d.h>

using namespace Eigen;
using namespace wolf;

/**
 *  SET TO TRUE  TO PRODUCE CSV FILE
 *  SET TO FALSE TO STOP PRODUCING CSV FILE
 */
#define WRITE_CSV_FILE  true

class ProcessorImuStaticInitTest : public testing::Test
{

    public: //These can be accessed in fixtures
        wolf::ProblemPtr problem_ptr_;
        wolf::SensorBasePtr sensor_ptr_;
        wolf::ProcessorMotionPtr processor_motion_;
        wolf::TimeStamp t;
        wolf::TimeStamp t0;
        double dt;
        Vector6d data;
        Matrix6d data_cov;
        FrameBasePtr KF0_;
        FrameBasePtr first_frame_;
        FrameBasePtr last_frame_;
        SolverCeresPtr solver_;

    //a new of this will be created for each new test
    void SetUp() override
    {
        WOLF_INFO("Doing setup...");
        using namespace Eigen;
        using std::shared_ptr;
        using std::make_shared;
        using std::static_pointer_cast;
        using namespace wolf::Constants;

        std::string wolf_root = _WOLF_IMU_ROOT_DIR;

        // Wolf problem
        problem_ptr_ = Problem::create("POV", 3);
        Vector7d extrinsics = (Vector7d() << 0,0,0,0,0,0,1).finished();
        sensor_ptr_ = problem_ptr_->installSensor("SensorImu", "Main Imu", extrinsics,  wolf_root + "/test/yaml/sensor_imu_static_init.yaml");
        ProcessorBasePtr processor_ptr = problem_ptr_->installProcessor("ProcessorImu", "Imu pre-integrator", "Main Imu", wolf_root + "/test/yaml/imu_static_init.yaml");
        processor_motion_ = std::static_pointer_cast<ProcessorMotion>(processor_ptr);

        // Time and data variables
        dt = 0.1;
        t0.set(0);
        t = t0;
        data = Vector6d::Random();
        data_cov = Matrix6d::Identity();
        last_frame_ = nullptr;
        first_frame_ = nullptr;

        // Create the solver
        solver_ = make_shared<SolverCeres>(problem_ptr_);
        
        // Set the origin
        VectorComposite x0c("POV", {Vector3d::Zero(), (Vector4d() << 0,0,0,1).finished(), Vector3d::Zero()});
        WOLF_INFO("x0c is: \n", x0c);
        //KF0_ = problem_ptr_->setPriorFix(x0c, t0);

        Vector4d q_init; q_init << 0,0,0,1;
        VectorComposite x_origin("POV", {Vector3d::Zero(), q_init, Vector3d::Zero()});
        VectorComposite std_origin("POV", {0.001*Vector3d::Ones(), 0.001*Vector3d::Ones(), 0.001*Vector3d::Ones()});
        KF0_ = problem_ptr_->setPriorFactor(x_origin, std_origin, 0);

    }

    void TestStatic(const Vector6d& body_magnitudes, const Vector6d& bias_groundtruth, const Vector6d& bias_initial_guess, const string& test_name, int testing_var, bool small_tol)
    {
      //Data
      data = body_magnitudes;
      data.head(3) -= Quaterniond(Vector4d(KF0_->getO()->getState())).conjugate() * wolf::gravity();
      data += bias_groundtruth;

      //Set bias initial guess
      sensor_ptr_->getIntrinsic(t0)->setState(bias_initial_guess);
      processor_motion_->setOrigin(KF0_);

#if WRITE_CSV_FILE
      std::fstream file_est; 
      file_est.open("./est-"+test_name+".csv", std::fstream::out);
      //    std::string header_est = "t,px,py,pz,qx,qy,qz,qw,vx,vy,vz,bax_est,bax_preint\n";
      std::string header_est;
      if(testing_var == 0) header_est = "t;px;vx;ox;bax_est;bgx_est;bax_preint;bgx_preint\n";
      if(testing_var == 1) header_est = "t;py;vy;oy;bay_est;bgy_est;bay_preint;bgy_preint\n";
      if(testing_var == 2) header_est = "t;pz;vz;oz:baz_est;bgz_est;baz_preint;bgz_preint\n";
      if(testing_var == 3) header_est = "t;pnorm;vnorm;onorm;banorm_est;bgnorm_est;banorm_preint;bgnorm_preint\n";
      file_est << header_est;
#endif


      int n_frames = 0;
      for(t = t0; t <= t0 + 9.9 + dt/2; t+=dt){
        WOLF_INFO("\n------------------------------------------------------------------------");

        //Create and process capture  
        auto C = std::make_shared<CaptureImu>(t, sensor_ptr_, data, data_cov, KF0_->getCaptureList().front());
        C->process();

        auto state = problem_ptr_->getState();
        auto bias_est = sensor_ptr_->getIntrinsic()->getState();
        auto bias_preint = processor_motion_->getLast()->getCalibrationPreint();

#if WRITE_CSV_FILE
        // pre-solve print to CSV
        if(testing_var == 3){
        file_est << std::fixed << t << ";"
            << state['P'].norm() << ";"
            << state['V'].norm() << ";"
            << state['O'].norm() << ";"
            << bias_est.head(3).norm() << ";"
            << bias_est.tail(3).norm() << ";"
            << bias_preint.head(3).norm() << ";"
            << bias_preint.tail(3).norm() << "\n";
        }
        else
        {
        file_est << std::fixed << t << ";"
            << state['P'](testing_var) << ";"
            << state['V'](testing_var) << ";"
            << state['O'](testing_var) << ";"
            << bias_est(testing_var) << ";"
            << bias_est(testing_var+3) << ";"
            << bias_preint(testing_var) << ";"
            << bias_preint(testing_var+3) << "\n";

        }
#endif

        // new frame
        if (last_frame_ != processor_motion_->getOrigin()->getFrame())
        {
          n_frames++;
          last_frame_ = processor_motion_->getOrigin()->getFrame();

          // impose static
          last_frame_->getP()->setState(KF0_->getP()->getState());
          last_frame_->getO()->setState(KF0_->getO()->getState());
          last_frame_->getV()->setZero();

          //Fix frame
          last_frame_->fix();

          //Solve
          if(n_frames > 0)
          {
            std::string report = solver_->solve(SolverManager::ReportVerbosity::FULL);
            //WOLF_INFO("Solver Report: ", report);

            state = problem_ptr_->getState();
            bias_est = sensor_ptr_->getIntrinsic()->getState();
            bias_preint = processor_motion_->getLast()->getCalibrationPreint();
            
            WOLF_INFO("The current problem is:");
            problem_ptr_->print(4);

#if WRITE_CSV_FILE
            // post-solve print to CSV with time-stamp shifted by dt/2 to separate from pre-solve  result
            if(testing_var == 3){
              file_est << std::fixed << t+dt/2 << ";"
                << state['P'].norm() << ";"
                << state['V'].norm() << ";"
                << state['O'].norm() << ";"
                << bias_est.head(3).norm() << ";"
                << bias_est.tail(3).norm() << ";"
                << bias_preint.head(3).norm() << ";"
                << bias_preint.tail(3).norm() << "\n";
            }
            else
            {
              file_est << std::fixed << t+dt/2 << ";"
                << state['P'](testing_var) << ";"
                << state['V'](testing_var) << ";"
                << state['O'](testing_var) << ";"
                << bias_est(testing_var) << ";"
                << bias_est(testing_var+3) << ";"
                << bias_preint(testing_var) << ";"
                << bias_preint(testing_var+3) << "\n";

            }
#endif
          }

        }



        WOLF_INFO("Number of frames is ", n_frames);
        WOLF_INFO("The state is: ", state);
        WOLF_INFO("The estimated bias is: ", bias_est.transpose());
        WOLF_INFO("The preintegrated bias is: ", bias_preint.transpose());
        if(small_tol)
        {
          if(n_frames == 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-6);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-6);
          }
          else if (n_frames > 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-6);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-6);
            EXPECT_MATRIX_APPROX(bias_preint, bias_groundtruth, 1e-6);
          }
        }
        else
        {
          if(n_frames == 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-3);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-3);
          }
          else if (n_frames > 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-3);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-3);
            EXPECT_MATRIX_APPROX(bias_preint, bias_groundtruth, 1e-3);
          }
        }
      }

#if WRITE_CSV_FILE
      file_est.close();
#endif

    }

    void TestStaticZeroOdom(const Vector6d& body_magnitudes, const Vector6d& bias_groundtruth, const Vector6d& bias_initial_guess, const string& test_name, int testing_var, bool small_tol)
    {
      //Data
      data = body_magnitudes;
      data.head(3) -= Quaterniond(Vector4d(KF0_->getO()->getState())).conjugate() * wolf::gravity();
      data += bias_groundtruth;

      //Set bias initial guess
      sensor_ptr_->getIntrinsic(t0)->setState(bias_initial_guess);
      processor_motion_->setOrigin(KF0_);

#if WRITE_CSV_FILE
      std::fstream file_est; 
      file_est.open("./estzeroodom-"+test_name+".csv", std::fstream::out);
      //    std::string header_est = "t,px,py,pz,qx,qy,qz,qw,vx,vy,vz,bax_est,bax_preint\n";
      std::string header_est;
      if(testing_var == 0) header_est = "t;px;vx;ox;bax_est;bgx_est;bax_preint;bgx_preint\n";
      if(testing_var == 1) header_est = "t;py;vy;oy;bay_est;bgy_est;bay_preint;bgy_preint\n";
      if(testing_var == 2) header_est = "t;pz;vz;oz:baz_est;bgz_est;baz_preint;bgz_preint\n";
      if(testing_var == 3) header_est = "t;pnorm;vnorm;onorm;banorm_est;bgnorm_est;banorm_preint;bgnorm_preint\n";
      file_est << header_est;
#endif


      int n_frames = 0;
      for(t = t0; t <= t0 + 9.9 + dt/2; t+=dt){
        WOLF_INFO("\n------------------------------------------------------------------------");

        //Create and process capture  
        auto C = std::make_shared<CaptureImu>(t, sensor_ptr_, data, data_cov, KF0_->getCaptureList().front());
        C->process();

        auto state = problem_ptr_->getState();
        auto bias_est = sensor_ptr_->getIntrinsic()->getState();
        auto bias_preint = processor_motion_->getLast()->getCalibrationPreint();

#if WRITE_CSV_FILE
        // pre-solve print to CSV
        if(testing_var == 3){
        file_est << std::fixed << t << ";"
            << state['P'].norm() << ";"
            << state['V'].norm() << ";"
            << state['O'].norm() << ";"
            << bias_est.head(3).norm() << ";"
            << bias_est.tail(3).norm() << ";"
            << bias_preint.head(3).norm() << ";"
            << bias_preint.tail(3).norm() << "\n";
        }
        else
        {
        file_est << std::fixed << t << ";"
            << state['P'](testing_var) << ";"
            << state['V'](testing_var) << ";"
            << state['O'](testing_var) << ";"
            << bias_est(testing_var) << ";"
            << bias_est(testing_var+3) << ";"
            << bias_preint(testing_var) << ";"
            << bias_preint(testing_var+3) << "\n";

        }
#endif

        // new frame
        if (last_frame_ != processor_motion_->getOrigin()->getFrame())
        {
          n_frames++;
          last_frame_ = processor_motion_->getOrigin()->getFrame();
          
          // impose zero odometry
          processor_motion_->setOdometry(sensor_ptr_->getProblem()->stateZero(processor_motion_->getStateStructure()));

          // add zero displacement and rotation capture & feature & factor with all previous frames
          assert(sensor_ptr_->getProblem());
          for (auto frm_pair =  sensor_ptr_->getProblem()->getTrajectory()->begin();
              frm_pair != sensor_ptr_->getProblem()->getTrajectory()->end();
              frm_pair++)
          {
            if (frm_pair->second == last_frame_)
              break;
            auto capture_zero = CaptureBase::emplace<CaptureVoid>(last_frame_, last_frame_->getTimeStamp(), nullptr);

            auto feature_zero = FeatureBase::emplace<FeatureBase>(capture_zero,
                "FeatureZeroOdom",
                Vector7d::Zero(),
                Eigen::MatrixXd::Identity(6,6) * 0.01);

            FactorBase::emplace<FactorRelativePose3d>(feature_zero,
                feature_zero,
                frm_pair->second,
                nullptr,
                false,
                TOP_MOTION);

          }

          // impose static
          last_frame_->getV()->setZero();

          //Fix frame
          last_frame_->getV()->fix();

          //Solve
          if(n_frames > 0)
          {
            std::string report = solver_->solve(SolverManager::ReportVerbosity::FULL);
            //WOLF_INFO("Solver Report: ", report);

            state = problem_ptr_->getState();
            bias_est = sensor_ptr_->getIntrinsic()->getState();
            bias_preint = processor_motion_->getLast()->getCalibrationPreint();
            
            WOLF_INFO("The current problem is:");
            problem_ptr_->print(4);

#if WRITE_CSV_FILE
            // post-solve print to CSV with time-stamp shifted by dt/2 to separate from pre-solve  result
            if(testing_var == 3){
              file_est << std::fixed << t+dt/2 << ";"
                << state['P'].norm() << ";"
                << state['V'].norm() << ";"
                << state['O'].norm() << ";"
                << bias_est.head(3).norm() << ";"
                << bias_est.tail(3).norm() << ";"
                << bias_preint.head(3).norm() << ";"
                << bias_preint.tail(3).norm() << "\n";
            }
            else
            {
              file_est << std::fixed << t+dt/2 << ";"
                << state['P'](testing_var) << ";"
                << state['V'](testing_var) << ";"
                << state['O'](testing_var) << ";"
                << bias_est(testing_var) << ";"
                << bias_est(testing_var+3) << ";"
                << bias_preint(testing_var) << ";"
                << bias_preint(testing_var+3) << "\n";

            }
#endif
          }

        }



        WOLF_INFO("Number of frames is ", n_frames);
        WOLF_INFO("The state is: ", state);
        WOLF_INFO("The estimated bias is: ", bias_est.transpose());
        WOLF_INFO("The preintegrated bias is: ", bias_preint.transpose());
        if(small_tol)
        {
          if(n_frames == 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-6);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-6);
          }
          else if (n_frames > 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-6);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-6);
            EXPECT_MATRIX_APPROX(bias_preint, bias_groundtruth, 1e-6);
          }
        }
        else
        {
          if(n_frames == 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-3);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-3);
          }
          else if (n_frames > 2)
          {
            EXPECT_MATRIX_APPROX(state.vector("POV"), KF0_->getState().vector("POV"), 1e-3);
            EXPECT_MATRIX_APPROX(bias_est, bias_groundtruth, 1e-3);
            EXPECT_MATRIX_APPROX(bias_preint, bias_groundtruth, 1e-3);
          }
        }
      }

#if WRITE_CSV_FILE
      file_est.close();
#endif

    }
};



TEST_F(ProcessorImuStaticInitTest, static_bias_aX_initial_guess_zero)
{
   Vector6d body_magnitudes = Vector6d::Zero();
   Vector6d bias_groundtruth = (Vector6d() << 0.42, 0, 0, 0, 0, 0).finished(); 
   Vector6d bias_initial_guess = Vector6d::Zero();

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_aX_initial_guess_zero", 0, true); 
}

TEST_F(ProcessorImuStaticInitTest, static_bias_zero_initial_guess_aX)
{
   Vector6d body_magnitudes = Vector6d::Zero();
   Vector6d bias_groundtruth = Vector6d::Zero();
   Vector6d bias_initial_guess = (Vector6d() << 0.42, 0, 0, 0, 0, 0).finished(); 

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_aX", 0, true); 
}

TEST_F(ProcessorImuStaticInitTest, static_bias_gX_initial_guess_zero)
{
   Vector6d body_magnitudes = Vector6d::Zero();
   Vector6d bias_groundtruth = (Vector6d() << 0, 0, 0, 0.01, 0, 0).finished(); 
   Vector6d bias_initial_guess = Vector6d::Zero();

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_gX_initial_guess_zero", 0, false); 
}

TEST_F(ProcessorImuStaticInitTest, static_bias_zero_initial_guess_gX)
{
   Vector6d body_magnitudes = Vector6d::Zero();
   Vector6d bias_groundtruth = Vector6d::Zero();
   Vector6d bias_initial_guess = (Vector6d() << 0, 0, 0, 0.01, 0, 0).finished(); 

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_gX", 0, false); 
}

TEST_F(ProcessorImuStaticInitTest, static_bias_aX_initial_guess_zero_zeroodom)
{
   Vector6d body_magnitudes = Vector6d::Zero();
   Vector6d bias_groundtruth = (Vector6d() << 0.42, 0, 0, 0, 0, 0).finished(); 
   Vector6d bias_initial_guess = Vector6d::Zero();

   TestStaticZeroOdom(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_aX_initial_guess_zero", 0, true); 
}

TEST_F(ProcessorImuStaticInitTest, static_bias_zero_initial_guess_aX_zeroodom)
{
   Vector6d body_magnitudes = Vector6d::Zero();
   Vector6d bias_groundtruth = Vector6d::Zero();
   Vector6d bias_initial_guess = (Vector6d() << 0.42, 0, 0, 0, 0, 0).finished(); 

   TestStaticZeroOdom(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_aX", 0, true); 
}

TEST_F(ProcessorImuStaticInitTest, static_bias_gX_initial_guess_zero_zeroodom)
{
   Vector6d body_magnitudes = Vector6d::Zero();
   Vector6d bias_groundtruth = (Vector6d() << 0, 0, 0, 0.01, 0, 0).finished(); 
   Vector6d bias_initial_guess = Vector6d::Zero();

   TestStaticZeroOdom(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_gX_initial_guess_zero", 0, false); 
}

TEST_F(ProcessorImuStaticInitTest, static_bias_zero_initial_guess_gX_zeroodom)
{
   Vector6d body_magnitudes = Vector6d::Zero();
   Vector6d bias_groundtruth = Vector6d::Zero();
   Vector6d bias_initial_guess = (Vector6d() << 0, 0, 0, 0.01, 0, 0).finished(); 

   TestStaticZeroOdom(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_gX", 0, false); 
}


TEST_F(ProcessorImuStaticInitTest, static_bias_zero_initial_guess_a_random)
{
   Vector6d body_magnitudes = Vector6d::Zero();
   Vector6d bias_groundtruth = Vector6d::Zero();
   Vector6d bias_initial_guess = Vector6d::Random()*100;
   bias_initial_guess.tail(3) = Vector3d::Zero();

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_a_random", 3, true); 
}

TEST_F(ProcessorImuStaticInitTest, static_bias_zero_initial_guess_random)
{
   Vector6d body_magnitudes = Vector6d::Zero();
   Vector6d bias_groundtruth = Vector6d::Zero();
   Vector6d bias_initial_guess = Vector6d::Random()*0.01;

   TestStatic(body_magnitudes, bias_groundtruth, bias_initial_guess, "static_bias_zero_initial_guess_random", 3, false); 
}


int main(int argc, char **argv)
{
    testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
