//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/**
 * \file gtest_processor_imu2d.cpp
 *
 *  Created on: Nov 24, 2020
 *      \author: igeer
 */

#include "imu/internal/config.h"
#include "imu/capture/capture_imu.h"
#include "imu/processor/processor_imu2d.h"
#include "imu/sensor/sensor_imu2d.h"

// #include "core/common/wolf.h"

#include <core/utils/utils_gtest.h>
#include "core/utils/logging.h"

#include "core/math/rotations.h"
#include <cmath>
#include <iostream>

using namespace Eigen;
using namespace wolf;

class ProcessorImu2dTest : public testing::Test
{

    public: //These can be accessed in fixtures
        wolf::ProblemPtr problem;
        wolf::SensorBasePtr sensor_ptr;
        wolf::ProcessorMotionPtr processor_motion;
        wolf::TimeStamp t;
        wolf::TimeStamp t0;
        double mti_clock, tmp;
        double dt;
        Vector6d data;
        Matrix6d data_cov;
        VectorComposite     x0c;                                // initial state composite
        VectorComposite     s0c;                                // initial sigmas composite
        std::shared_ptr<wolf::CaptureImu> C;

    //a new of this will be created for each new test
    void SetUp() override
    {
        using namespace Eigen;
        using std::shared_ptr;
        using std::make_shared;
        using std::static_pointer_cast;
        using namespace wolf::Constants;

        std::string wolf_root = _WOLF_IMU_ROOT_DIR;

        // Wolf problem
        problem = Problem::create("POV", 2);
        Vector3d extrinsics = (Vector3d() << 0,0, 0).finished();
        sensor_ptr = problem->installSensor("SensorImu2d", "Main Imu", extrinsics,  wolf_root + "/demos/sensor_imu2d.yaml");
        ProcessorBasePtr processor_ptr = problem->installProcessor("ProcessorImu2d", "Imu pre-integrator", "Main Imu", wolf_root + "/demos/processor_imu2d.yaml");
        processor_motion = std::static_pointer_cast<ProcessorMotion>(processor_ptr);

        // Time and data variables
        dt = 0.01;
        t0.set(0);
        t = t0;
        data = Vector6d::Zero();
        data_cov = Matrix6d::Identity();

        // Create one capture to store the Imu data arriving from (sensor / callback / file / etc.)
        C = make_shared<CaptureImu>(t, sensor_ptr, data, data_cov, Vector3d::Zero());
    }

};

TEST(ProcessorImu2d_constructors, ALL)
{
    //constructor with ProcessorImu2dParamsPtr argument only
    ParamsProcessorImu2dPtr param_ptr = std::make_shared<ParamsProcessorImu2d>();
    param_ptr->max_time_span =   2.0;
    param_ptr->max_buff_length = 20000;
    param_ptr->dist_traveled =   2.0;
    param_ptr->angle_turned =    2.0;

    ProcessorImu2dPtr prc1 = std::make_shared<ProcessorImu2d>(param_ptr);
    ASSERT_EQ(prc1->getMaxTimeSpan(), param_ptr->max_time_span);
    ASSERT_EQ(prc1->getMaxBuffLength(), param_ptr->max_buff_length);
    ASSERT_EQ(prc1->getDistTraveled(), param_ptr->dist_traveled);
    ASSERT_EQ(prc1->getAngleTurned(), param_ptr->angle_turned);

    //Factory constructor with pointers
    std::string wolf_root = _WOLF_IMU_ROOT_DIR;
    ProblemPtr problem = Problem::create("POV", 2);
    Vector3d extrinsics = (Vector3d()<<1,0, 0).finished();
    std::cout << "creating sensor_ptr" << std::endl;
    SensorBasePtr sensor_ptr = problem->installSensor("SensorImu2d", "Main Imu", extrinsics, wolf_root + "/demos/sensor_imu2d.yaml");
    std::cout << "created sensor_ptr" << std::endl;
    ParamsProcessorImu2dPtr params_default = std::make_shared<ParamsProcessorImu2d>();
    ProcessorBasePtr processor_ptr = problem->installProcessor("ProcessorImu2d", "Imu pre-integrator", sensor_ptr, params_default);
    ASSERT_EQ(std::static_pointer_cast<ProcessorImu2d>(processor_ptr)->getMaxTimeSpan(),   params_default->max_time_span);
    ASSERT_EQ(std::static_pointer_cast<ProcessorImu2d>(processor_ptr)->getMaxBuffLength(), params_default->max_buff_length);
    ASSERT_EQ(std::static_pointer_cast<ProcessorImu2d>(processor_ptr)->getDistTraveled(),  params_default->dist_traveled);
    ASSERT_EQ(std::static_pointer_cast<ProcessorImu2d>(processor_ptr)->getAngleTurned(),   params_default->angle_turned);

    //Factory constructor with yaml
    processor_ptr = problem->installProcessor("ProcessorImu2d", "Sec Imu pre-integrator", "Main Imu", wolf_root + "/demos/processor_imu2d.yaml");
    ASSERT_EQ(std::static_pointer_cast<ProcessorImu2d>(processor_ptr)->getMaxTimeSpan(),   2.0);
    ASSERT_EQ(std::static_pointer_cast<ProcessorImu2d>(processor_ptr)->getMaxBuffLength(), 20000);
    ASSERT_EQ(std::static_pointer_cast<ProcessorImu2d>(processor_ptr)->getDistTraveled(),  2.0);
    ASSERT_EQ(std::static_pointer_cast<ProcessorImu2d>(processor_ptr)->getAngleTurned(),   0.2);
}

TEST_F(ProcessorImu2dTest, Prior)
{
    // Set the origin
    x0c['P'] = Vector2d(1,2);
    x0c['O'] = Vector1d(0);
    x0c['V'] = Vector2d(4,5);
    auto KF0 = problem->setPriorFix(x0c, t0);
    processor_motion->setOrigin(KF0);
    //WOLF_DEBUG("x0 =", x0c);
    //WOLF_DEBUG("KF0 state =", problem->getTrajectory()->getFrameMap().at(t)->getState("POV"));
}

TEST_F(ProcessorImu2dTest, MRUA)
{
   data << 1, 0, 0,   0, 0, 0; 
   data_cov *= 1e-3;
   // Set the origin
   x0c['P'] = Vector2d(1,2);
   x0c['O'] = Vector1d(0);
   x0c['V'] = Vector2d(4,5);
   auto KF0 = problem->setPriorFix(x0c, t0);
   processor_motion->setOrigin(KF0);

   //WOLF_DEBUG("Current State: ", problem->getState());
   for(t = t0+dt; t <= t0 + 1.0 + dt/2; t+=dt){
       C->setTimeStamp(t);
       C->setData(data);
       C->setDataCovariance(data_cov);
       C->process();
       //WOLF_DEBUG("Current State: ", problem->getState()['P'].transpose());
       //WOLF_DEBUG((x0c['P'] + x0c['V']*(t-t0) + 0.5*data.head(2)*std::pow(t-t0, 2)).transpose());
       ASSERT_MATRIX_APPROX(x0c['P'] + x0c['V']*(t-t0) + 0.5*data.head(2)*std::pow(t-t0, 2), problem->getState()['P'], 1e-9);
   }
}

TEST_F(ProcessorImu2dTest, parabola)
{
   double v0 = 10;
   double a = 1;

   Vector2d pos;
   // Set the origin
   x0c['P'] = Vector2d(0, 0);
   x0c['O'] = Vector1d(0);
   x0c['V'] = Vector2d(v0, 0);

   data_cov *= 1e-3;
   auto KF0 = problem->setPriorFix(x0c, t0);
   processor_motion->setOrigin(KF0);

   for(t = t0+dt; t <= t0 + 0.5 + dt/2; t+=dt){
       C->setTimeStamp(t);
       data <<  0, a, 0,    0,  0, 0;
       C->setData(data);
       C->setDataCovariance(data_cov);
       C->process();
       //WOLF_DEBUG("Current State: ", problem->getState());
       pos << v0*(t-t0),
              0.5*a*std::pow(t-t0, 2);
       //WOLF_DEBUG("Calculated State: ", pos.transpose());
       EXPECT_MATRIX_APPROX(pos , problem->getState()['P'], 1e-9);
       EXPECT_MATRIX_APPROX(Vector2d(v0, a*(t-t0)), problem->getState()['V'], 1e-9);
   }
}

TEST_F(ProcessorImu2dTest, parabola_deluxe)
{
   Vector2d v0(13, 56);
   Vector2d a(1, 2);

   Vector2d pos;
   // Set the origin
   x0c['P'] = Vector2d(0, 0);
   x0c['O'] = Vector1d(0);
   x0c['V'] = v0;

   data_cov *= 1e-3;
   auto KF0 = problem->setPriorFix(x0c, t0);
   processor_motion->setOrigin(KF0);

   for(t = t0+dt; t <= t0 + 0.5 + dt/2; t+=dt){
       C->setTimeStamp(t);
       data <<  a(0), a(1), 0,    0,  0, 0;
       C->setData(data);
       C->setDataCovariance(data_cov);
       C->process();
       //WOLF_DEBUG("Current State: ", problem->getState());
       pos = v0*(t-t0) + 0.5*a*std::pow(t-t0, 2);
       //WOLF_DEBUG("Calculated State: ", pos.transpose());
       EXPECT_MATRIX_APPROX(pos , problem->getState()['P'], 1e-9);
       EXPECT_MATRIX_APPROX(v0 + a*(t-t0), problem->getState()['V'], 1e-9);
   }
}

TEST_F(ProcessorImu2dTest, Circular_move)
{
   double pi = 3.14159265358993;
   double r = 1;
   double w = 1;
   double alpha = pi/4.;
   Vector2d pos;
   // Set the origin
   x0c['P'] = Vector2d(r, 0);
   pos = x0c['P'];
   x0c['O'] = Vector1d(alpha);
   x0c['V'] = Vector2d(0, w*r);

   data_cov *= 1e-3;
   //dt = 0.0001;
   auto KF0 = problem->setPriorFix(x0c, t0);
   processor_motion->setOrigin(KF0);

   WOLF_DEBUG("Current State: ", problem->getState());
   for(t = t0+dt; t <= t0 + 0.5 + dt/2; t+=dt){
       C->setTimeStamp(t);
       data <<  -std::cos(alpha)*w*w*r, std::sin(alpha)*w*w*r, 0,    0,  0, w;
       C->setData(data);
       C->setDataCovariance(data_cov);
       C->process();
       WOLF_DEBUG("Current State: ", problem->getState());
       pos << r*std::cos( w*(t-t0) ),
              r*std::sin( w*(t-t0) );
       WOLF_DEBUG("Calculated State: ", pos.transpose());
       EXPECT_MATRIX_APPROX(pos , problem->getState()['P'], 1e-9);
   }
}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  //::testing::GTEST_FLAG(filter) = "ProcessorImu2dt.check_Covariance";
  return RUN_ALL_TESTS();
}

