//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include "core/utils/utils_gtest.h"

#include "imu/internal/config.h"
#include "imu/capture/capture_compass.h"
#include "imu/sensor/sensor_compass.h"

#include "core/math/covariance.h"

#include <cstdio>

using namespace wolf;
using namespace Eigen;

std::string wolf_root = _WOLF_IMU_ROOT_DIR;

TEST(SensorCompass, constructor)
{
    auto params = std::make_shared<ParamsSensorCompass>();
    Vector7d extr(Vector7d::Random());
    extr.tail<4>().normalize();

    ASSERT_NEAR(extr.tail<4>().norm(), 1, 1e-6);

    auto sen = std::make_shared<SensorCompass>(extr, params);

    ASSERT_NE(sen, nullptr);

    ASSERT_MATRIX_APPROX(sen->getP()->getState(), extr.head(3), 1e-12);
    ASSERT_MATRIX_APPROX(sen->getO()->getState(), extr.tail(4), 1e-12);
}

TEST(SensorCompass, getParams)
{
    auto params = std::make_shared<ParamsSensorCompass>();
    params->stdev_noise = 0.1;
    params->field_prior_mode = "fix";
    params->field_prior_state = Eigen::Vector3d(1, 2, 3);
    params->field_prior_sigma = Eigen::Vector3d(4, 5, 6);

    params->bias_prior_mode = "factor";
    params->bias_prior_state = Eigen::Vector3d(3, 2, 1);
    params->bias_prior_sigma = Eigen::Vector3d(6, 5, 4);

    Vector7d extr(Vector7d::Random());
    extr.tail<4>().normalize();

    ASSERT_NEAR(extr.tail<4>().norm(), 1, 1e-6);

    auto sen = std::make_shared<SensorCompass>(extr, params);

    ASSERT_NE(sen->getParams(), nullptr);

    ASSERT_NEAR(sen->getParams()->stdev_noise, 0.1, 1e-9);

    ASSERT_EQ(sen->getParams()->field_prior_mode, "fix");
    ASSERT_MATRIX_APPROX(sen->getParams()->field_prior_state, Eigen::Vector3d(1, 2, 3), 1e-9);
    ASSERT_MATRIX_APPROX(sen->getParams()->field_prior_sigma, Eigen::Vector3d(4, 5, 6), 1e-9);

    ASSERT_EQ(sen->getParams()->bias_prior_mode, "factor");
    ASSERT_MATRIX_APPROX(sen->getParams()->bias_prior_state, Eigen::Vector3d(3, 2, 1), 1e-9);
    ASSERT_MATRIX_APPROX(sen->getParams()->bias_prior_sigma, Eigen::Vector3d(6, 5, 4), 1e-9);

    ASSERT_FALSE(sen->getIntrinsic()->isFixed());
    ASSERT_TRUE(sen->getStateBlock('F')->isFixed());
}

TEST(SensorCompass, create)
{
    auto params = std::make_shared<ParamsSensorCompass>();
    params->stdev_noise = 0.1;
    params->field_prior_mode = "fix";
    params->field_prior_state = Eigen::Vector3d(1, 2, 3);
    params->field_prior_sigma = Eigen::Vector3d(4, 5, 6);

    params->bias_prior_mode = "factor";
    params->bias_prior_state = Eigen::Vector3d(3, 2, 1);
    params->bias_prior_sigma = Eigen::Vector3d(6, 5, 4);

    Vector7d extr(Vector7d::Random());
    extr.tail<4>().normalize();

    auto sen_base = SensorCompass::create("name", extr, params);

    auto sen = std::static_pointer_cast<SensorCompass>(sen_base);

    ASSERT_NE(sen->getParams(), nullptr);

    ASSERT_NEAR(sen->getParams()->stdev_noise, 0.1, 1e-9);

    ASSERT_EQ(sen->getParams()->field_prior_mode, "fix");
    ASSERT_MATRIX_APPROX(sen->getParams()->field_prior_state, Eigen::Vector3d(1, 2, 3), 1e-9);
    ASSERT_MATRIX_APPROX(sen->getParams()->field_prior_sigma, Eigen::Vector3d(4, 5, 6), 1e-9);

    ASSERT_EQ(sen->getParams()->bias_prior_mode, "factor");
    ASSERT_MATRIX_APPROX(sen->getParams()->bias_prior_state, Eigen::Vector3d(3, 2, 1), 1e-9);
    ASSERT_MATRIX_APPROX(sen->getParams()->bias_prior_sigma, Eigen::Vector3d(6, 5, 4), 1e-9);

    ASSERT_FALSE(sen->getIntrinsic()->isFixed());
    ASSERT_TRUE(sen->getStateBlock('F')->isFixed());
}

TEST(SensorCompass, install)
{
    auto params = std::make_shared<ParamsSensorCompass>();
    params->stdev_noise = 0.1;
    params->field_prior_mode = "fix";
    params->field_prior_state = Eigen::Vector3d(1, 2, 3);
    params->field_prior_sigma = Eigen::Vector3d(4, 5, 6);

    params->bias_prior_mode = "factor";
    params->bias_prior_state = Eigen::Vector3d(3, 2, 1);
    params->bias_prior_sigma = Eigen::Vector3d(6, 5, 4);

    Vector7d extr(Vector7d::Random());
    extr.tail<4>().normalize();

    // Problem
    ProblemPtr problem_ptr = Problem::create("PO", 3);

    // Install sensor
    auto sen_base = problem_ptr->installSensor("SensorCompass",
                                               "magnetometer",
                                               extr,
                                               params);

    auto sen = std::static_pointer_cast<SensorCompass>(sen_base);

    ASSERT_NE(sen->getParams(), nullptr);

    ASSERT_NEAR(sen->getParams()->stdev_noise, 0.1, 1e-9);

    ASSERT_EQ(sen->getParams()->field_prior_mode, "fix");
    ASSERT_MATRIX_APPROX(sen->getParams()->field_prior_state, Eigen::Vector3d(1, 2, 3), 1e-9);
    ASSERT_MATRIX_APPROX(sen->getParams()->field_prior_sigma, Eigen::Vector3d(4, 5, 6), 1e-9);

    ASSERT_EQ(sen->getParams()->bias_prior_mode, "factor");
    ASSERT_MATRIX_APPROX(sen->getParams()->bias_prior_state, Eigen::Vector3d(3, 2, 1), 1e-9);
    ASSERT_MATRIX_APPROX(sen->getParams()->bias_prior_sigma, Eigen::Vector3d(6, 5, 4), 1e-9);

    ASSERT_FALSE(sen->getIntrinsic()->isFixed());
    ASSERT_TRUE(sen->getStateBlock('F')->isFixed());
}

TEST(SensorCompass, computeMeasurementCovariance)
{
    auto params = std::make_shared<ParamsSensorCompass>();
    params->stdev_noise = 0.1;
    params->field_prior_mode = "fix";
    params->field_prior_state = Eigen::Vector3d(1, 2, 3);
    params->field_prior_sigma = Eigen::Vector3d(4, 5, 6);

    params->bias_prior_mode = "factor";
    params->bias_prior_state = Eigen::Vector3d(3, 2, 1);
    params->bias_prior_sigma = Eigen::Vector3d(6, 5, 4);

    Vector7d extr(Vector7d::Random());
    extr.tail<4>().normalize();

    auto sen_base = SensorCompass::create("name", extr, params);

    auto sen = std::static_pointer_cast<SensorCompass>(sen_base);

    for (auto i = 0; i < 3; i++)
    {
        Vector3d field(Vector3d::Zero());
        field(i) = 10;
        WOLF_INFO("field = ", field.transpose());

        Matrix3d cov = sen->computeMeasurementCovariance(field);
        WOLF_INFO("cov = \n", cov);

        Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d> eigensolver(cov, Eigen::ComputeEigenvectors);
        EXPECT_TRUE(eigensolver.info() == Eigen::Success);
        WOLF_INFO("eigenvalues: ", eigensolver.eigenvalues().transpose());
        WOLF_INFO("eigenvectors: \n", eigensolver.eigenvectors());

        // covariance is created from the direction of field:
        // large covariance: 1e6 m² in the field direction
        // small covariance: params->stdev_noise² in the other directions.

        // field should be aligned with its weighted vector by the covariance
        Vector3d field_w = cov * field;
        WOLF_INFO("field = ", field.transpose());
        WOLF_INFO("field_w = ", field_w.transpose());
        EXPECT_NEAR(field.norm(), field_w.norm() * 1e-6, 1e-6);
        EXPECT_LE(acos(field.dot(field_w) / (field.norm() * field_w.norm())), 0.1);
    }
}

TEST(SensorCompass, computeMeasurementCovarianceRandom)
{
    auto params = std::make_shared<ParamsSensorCompass>();
    params->stdev_noise = 0.1;
    params->field_prior_mode = "fix";
    params->field_prior_state = Eigen::Vector3d(1, 2, 3);
    params->field_prior_sigma = Eigen::Vector3d(4, 5, 6);

    params->bias_prior_mode = "factor";
    params->bias_prior_state = Eigen::Vector3d(3, 2, 1);
    params->bias_prior_sigma = Eigen::Vector3d(6, 5, 4);

    Vector7d extr(Vector7d::Random());
    extr.tail<4>().normalize();

    auto sen_base = SensorCompass::create("name", extr, params);

    auto sen = std::static_pointer_cast<SensorCompass>(sen_base);

    for (auto i = 0; i < 1e2; i++)
    {
        Vector3d field = Vector3d::Random().normalized() * 10;
        WOLF_INFO("field = ", field.transpose());

        Matrix3d cov = sen->computeMeasurementCovariance(field);
        WOLF_INFO("cov = \n", cov);

        Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d> eigensolver(cov, Eigen::ComputeEigenvectors);
        EXPECT_TRUE(eigensolver.info() == Eigen::Success);
        WOLF_INFO("eigenvalues: ", eigensolver.eigenvalues().transpose());
        WOLF_INFO("eigenvectors: \n", eigensolver.eigenvectors());

        // covariance is created from the direction of field:
        // large covariance: 1e6 m² in the field direction
        // small covariance: params->stdev_noise² in the other directions.

        // field should be aligned with its weighted vector by the covariance
        Vector3d field_w = cov * field;
        WOLF_INFO("field = ", field.transpose());
        WOLF_INFO("field_w = ", field_w.transpose());
        WOLF_INFO("field.norm() * field_w.norm() = ", field.norm() * field_w.norm());
        WOLF_INFO("field.dot(field_w) = ", field.dot(field_w));
        WOLF_INFO("field.dot(field_w) / (field.norm() * field_w.norm()) = ", field.dot(field_w) / (field.norm() * field_w.norm()));
        WOLF_INFO("acos(min(1.0,field.dot(field_w) / (field.norm() * field_w.norm()))) = ", acos(std::min(1.0,field.dot(field_w) / (field.norm() * field_w.norm()))));
        //EXPECT_NEAR(field.norm(), field_w.norm() * 1e-6, 1e-6);
        EXPECT_LE(acos(std::min(1.0,field.dot(field_w) / (field.norm() * field_w.norm()))), 0.1);
    }
}

int main(int argc, char **argv)
{
    testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}

